# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/4/2 16:18
# @Author  : wangchongshi
# @Email   : wangchongshi.wcs@antgroup.com
# @FileName: langchain_instance.py

from typing import Any, List, Optional, AsyncIterator

from langchain.callbacks.manager import AsyncCallbackManagerForLLMRun, CallbackManagerForLLMRun
from langchain_community.chat_models.openai import ChatOpenAI
from langchain.schema import BaseMessage, ChatResult
from langchain_community.chat_models.openai import _convert_delta_to_message_chunk
from langchain_core.language_models.chat_models import generate_from_stream, agenerate_from_stream
from langchain_core.messages import AIMessageChunk
from langchain_core.outputs import ChatGenerationChunk

from agentuniverse.llm.llm import LLM


class LangchainOpenAI(ChatOpenAI):
    """Langchain OpenAI LLM wrapper."""

    llm: Optional[LLM] = None

    def __init__(self, llm: LLM):
        """The __init__ method.

        The agentUniverse LLM instance is passed to this class as an argument.
        Convert the attributes of agentUniverse(aU) LLM instance to the LangchainOpenAI object for initialization

        Args:
            llm (LLM): the agentUniverse(aU) LLM instance.
        """
        init_params = dict()
        init_params['model_name'] = llm.model_name if llm.model_name else 'gpt-3.5-turbo'
        init_params['temperature'] = llm.temperature if llm.temperature else 0.7
        init_params['request_timeout'] = llm.request_timeout
        init_params['max_tokens'] = llm.max_tokens
        init_params['max_retries'] = llm.max_retries if llm.max_retries else 2
        init_params['streaming'] = llm.streaming if llm.streaming else False
        init_params['openai_api_key'] = llm.openai_api_key if llm.openai_api_key else 'blank'
        init_params['openai_organization'] = llm.openai_organization
        init_params['openai_api_base'] = llm.openai_api_base
        init_params['openai_proxy'] = llm.openai_proxy
        super().__init__(**init_params)
        self.llm = llm

    def _generate(
            self,
            messages: List[BaseMessage],
            stop: Optional[List[str]] = None,
            run_manager: Optional[CallbackManagerForLLMRun] = None,
            stream: Optional[bool] = None,
            **kwargs,
    ) -> ChatResult:
        """Run the Langchain OpenAI LLM."""
        should_stream = stream if stream is not None else self.streaming
        message_dicts, params = self._create_message_dicts(messages, stop)
        params = {**params, **kwargs}
        llm_output = self.llm.call(messages=message_dicts, **params)
        if not should_stream:
            return self._create_chat_result(llm_output.raw)
        stream_iter = self.as_langchain_chunk(llm_output)
        return generate_from_stream(stream_iter)

    async def _agenerate(
            self,
            messages: List[BaseMessage],
            stop: Optional[List[str]] = None,
            run_manager: Optional[AsyncCallbackManagerForLLMRun] = None,
            stream: Optional[bool] = None,
            **kwargs: Any,
    ) -> ChatResult:
        """Asynchronously run the Langchain OpenAI LLM."""
        should_stream = stream if stream is not None else self.streaming
        message_dicts, params = self._create_message_dicts(messages, stop)
        params = {**params, **kwargs}
        llm_output = await self.llm.acall(messages=message_dicts, **params)
        if not should_stream:
            return self._create_chat_result(llm_output.raw)
        stream_iter = self.as_langchain_achunk(llm_output)
        return await agenerate_from_stream(stream_iter)

    @staticmethod
    def as_langchain_chunk(stream, run_manager=None):
        default_chunk_class = AIMessageChunk
        for llm_result in stream:
            chunk = llm_result.raw
            if not isinstance(chunk, dict):
                chunk = chunk.dict()
            if len(chunk["choices"]) == 0:
                continue
            choice = chunk["choices"][0]
            chunk = _convert_delta_to_message_chunk(
                choice["delta"], default_chunk_class
            )
            finish_reason = choice.get("finish_reason")
            generation_info = (
                dict(finish_reason=finish_reason) if finish_reason is not None else None
            )
            default_chunk_class = chunk.__class__
            chunk = ChatGenerationChunk(message=chunk, generation_info=generation_info)
            yield chunk
            if run_manager:
                run_manager.on_llm_new_token(chunk.text, chunk=chunk)

    @staticmethod
    async def as_langchain_achunk(stream_iterator: AsyncIterator, run_manager=None) \
            -> AsyncIterator[ChatGenerationChunk]:
        default_chunk_class = AIMessageChunk
        async for llm_result in stream_iterator:
            chunk = llm_result.raw
            if not isinstance(chunk, dict):
                chunk = chunk.dict()
            if len(chunk["choices"]) == 0:
                continue
            choice = chunk["choices"][0]
            chunk = _convert_delta_to_message_chunk(
                choice["delta"], default_chunk_class
            )
            finish_reason = choice.get("finish_reason")
            generation_info = (
                dict(finish_reason=finish_reason) if finish_reason is not None else None
            )
            default_chunk_class = chunk.__class__
            chunk = ChatGenerationChunk(message=chunk, generation_info=generation_info)
            yield chunk
            if run_manager:
                await run_manager.on_llm_new_token(token=chunk.text, chunk=chunk)



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/8/20 19:18
# @Author  : wangchongshi
# @Email   : wangchongshi.wcs@antgroup.com
# @FileName: workflow.py
from typing import Optional

from agentuniverse.base.component.component_base import ComponentBase
from agentuniverse.base.component.component_enum import ComponentEnum
from agentuniverse.base.config.application_configer.application_config_manager import ApplicationConfigManager
from agentuniverse.base.config.component_configer.configers.workflow_configer import WorkflowConfiger
from agentuniverse.workflow.graph.graph import Graph
from agentuniverse.workflow.workflow_output import WorkflowOutput


class Workflow(ComponentBase):
    """The basic class of the workflow."""

    id: Optional[str] = None
    name: Optional[str] = None
    description: Optional[str] = None
    graph: Optional[Graph] = None
    graph_config: Optional[dict] = None

    class Config:
        """Configuration for this pydantic object."""
        arbitrary_types_allowed = True

    def __init__(self, **kwargs):
        super().__init__(component_type=ComponentEnum.WORKFLOW, **kwargs)

    def get_instance_code(self) -> str:
        """Return the full instance code of the workflow."""
        appname = ApplicationConfigManager().app_configer.base_info_appname
        return f'{appname}.{self.component_type.value.lower()}.{self.id}'

    def build(self) -> 'Workflow':
        if self.graph_config is None:
            raise ValueError('The graph config is None.')
        self.graph = Graph().build(self.id, self.graph_config)
        return self

    def run(self, input_params: dict) -> WorkflowOutput:
        if self.graph is None:
            raise ValueError('The graph of the workflow is None.')
        workflow_output = WorkflowOutput(workflow_id=self.id, workflow_start_params=input_params)
        self.graph.run(workflow_output)
        return workflow_output

    def initialize_by_component_configer(self, component_configer: WorkflowConfiger) -> 'Workflow':
        """Initialize the Workflow by the ComponentConfiger object.

        Args:
            component_configer(WorkflowConfiger): the ComponentConfiger object
        Returns:
            Workflow: the Workflow object
        """
        if component_configer.id:
            self.id = component_configer.id
        if component_configer.name:
            self.name = component_configer.name
        if component_configer.description:
            self.description = component_configer.description
        if component_configer.graph:
            self.graph_config = component_configer.graph
        return self



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/7/25 21:52
# @Author  : wangchongshi
# @Email   : wangchongshi.wcs@antgroup.com
# @FileName: agent_dto.py
from typing import Optional

from pydantic import BaseModel, Field

from agentuniverse_product.service.model.knowledge_dto import KnowledgeDTO
from agentuniverse_product.service.model.llm_dto import LlmDTO
from agentuniverse_product.service.model.planner_dto import PlannerDTO
from agentuniverse_product.service.model.prompt_dto import PromptDTO
from agentuniverse_product.service.model.tool_dto import ToolDTO


class AgentDTO(BaseModel):
    id: str = Field(description="ID")
    nickname: Optional[str] = Field(description="agent nickname", default="")
    avatar: Optional[str] = Field(description="agent avatar path", default="")
    description: Optional[str] = Field(description="agent description", default="")
    opening_speech: Optional[str] = Field(description="agent opening speech", default="")
    prompt: Optional[PromptDTO] = Field(description="agent prompt", default=None)
    llm: Optional[LlmDTO] = Field(description="agent llm", default=None)
    tool: Optional[list[ToolDTO]] = Field(description="agent tool list", default=[])
    memory: Optional[str] = Field(description="agent memory id", default='')
    planner: Optional[PlannerDTO] = Field(description="agent planner", default=None)
    knowledge: Optional[list[KnowledgeDTO]] = Field(description="agent knowledge list", default=[])
    mtime: Optional[float] = Field(description="product last modification time.", default=None)



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/10/17 11:09
# @Author  : wangchongshi
# @Email   : wangchongshi.wcs@antgroup.com
# @FileName: reviewing_agent.py
from agentuniverse.agent.template.reviewing_agent_template import ReviewingAgentTemplate


class ReviewingAgent(ReviewingAgentTemplate):
    """Reviewing Agent module."""



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/3/19 11:43
# @Author  : wangchongshi
# @Email   : wangchongshi.wcs@antgroup.com
# @FileName: openai_embedding.py

from typing import List, Optional, Any

from langchain_community.embeddings.openai import OpenAIEmbeddings
from openai import OpenAI, AsyncOpenAI, BadRequestError
from pydantic import Field

from agentuniverse.agent.action.knowledge.embedding.embedding import Embedding
from agentuniverse.base.util.env_util import get_from_env
from agentuniverse.base.config.component_configer.component_configer import \
    ComponentConfiger


class OpenAIEmbedding(Embedding):
    """The openai embedding class."""

    openai_client_args: Optional[dict] = None
    openai_api_key: Optional[str] = Field(default_factory=lambda: get_from_env("OPENAI_API_KEY"))
    client: Any = None
    async_client: Any = None
    dimensions: Optional[int] = None

    def get_embeddings(self, texts: List[str]) -> List[List[float]]:
        """Get the OpenAI embeddings.

        Note:
            The `embedding_model_name` parameter of the openai embedding class must be provided.
            The `dimensions` parameter of the openai embedding class is optional.

         Args:
             texts (List[str]): A list of texts that need to be embedded.

         Returns:
             List[List[float]]: Each text gets a float list, and the result is a list of the results for each text.

         Raises:
             ValueError: If texts exceed the embedding model token limit or missing some required parameters.
         """
        self.client = OpenAI(api_key=self.openai_api_key, **self.openai_client_args or {})
        if self.embedding_model_name is None:
            raise ValueError("Must provide `embedding_model_name`")
        try:
            if self.dimensions:
                response = self.client.embeddings.create(input=texts, model=self.embedding_model_name,
                                                         dimensions=self.dimensions)
            else:
                response = self.client.embeddings.create(input=texts, model=self.embedding_model_name)

            # Extract the embedding data from the response
            data = response.data

            # Return the embeddings as a list of lists of floats
            return [embedding.embedding for embedding in data]
        except BadRequestError as e:
            raise ValueError(e.message)

    async def async_get_embeddings(self, texts: List[str]) -> List[List[float]]:
        """Asynchronously get the OpenAI embeddings.

        Note:
            The `embedding_model_name` parameter of the openai embedding class must be provided.
            The `dimensions` parameter of the openai embedding class is optional.

         Args:
             texts (List[str]): A list of texts that need to be embedded.

         Returns:
             List[List[float]]: Each text gets a float list, and the result is a list of the results for each text.
         Raises:
             ValueError: If texts exceed the embedding model token limit or missing some required parameters.
         """
        self.async_client = AsyncOpenAI(api_key=self.openai_api_key, **self.openai_client_args or {})
        if self.embedding_model_name is None:
            raise ValueError("Must provide `embedding_model_name`")
        try:
            if self.dimensions:
                response = await self.async_client.embeddings.create(input=texts, model=self.embedding_model_name,
                                                                     dimensions=self.dimensions)
            else:
                response = await self.async_client.embeddings.create(input=texts, model=self.embedding_model_name)
            # Extract the embedding data from the response
            data = response.data

            # Return the embeddings as a list of lists of floats
            return [embedding.embedding for embedding in data]
        except BadRequestError as e:
            raise ValueError(e.message)

    def as_langchain(self) -> OpenAIEmbeddings:
        """Convert the agentUniverse(aU) openai embedding class to the langchain openai embedding class."""
        return OpenAIEmbeddings(openai_api_key=self.openai_api_key,
                                client=self.client.embeddings, async_client=self.async_client.embeddings)

    def _initialize_by_component_configer(self,
                                          embedding_configer: ComponentConfiger) \
            -> 'Embedding':
        """Initialize the embedding by the ComponentConfiger object.

        Args:
            embedding_configer(ComponentConfiger): A configer contains embedding
            basic info.
        Returns:
            Embedding: A embedding instance.
        """
        super()._initialize_by_component_configer(embedding_configer)
        if hasattr(embedding_configer, "embedding_dims"):
            self.dimensions = embedding_configer.embedding_dims
        if hasattr(embedding_configer, "openai_client_args"):
            self.openai_client_args = embedding_configer.openai_client_args
        return self



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/3/26 10:34
# @Author  : fanen.lhy
# @Email   : fanen.lhy@antgroup.com
# @FileName: web_util.py
import asyncio
import inspect
import queue
import json

from flask import request, make_response, jsonify

from ..service_instance import ServiceInstance
from ...agent.agent import Agent
from ...agent.agent_manager import AgentManager


def request_param(func):
    """An annotation used to parse the flask request params."""

    def wrapper(*args, **kwargs):
        if request.method == "GET":
            req_data = request.args.to_dict()
        # Get the post params from body according to different content type.
        else:
            if "application/json" in request.headers.get("Content-Type"):
                raw_data = request.data.decode('utf-8')
                req_data = json.loads(raw_data)
            else:
                req_data = request.form.to_dict()
        # Get the func arguments name and type.
        sig = inspect.signature(func)
        for name, param in sig.parameters.items():
            if name == "kwargs":
                for key in req_data:
                    if key not in kwargs:
                        kwargs[key] = req_data[key]
                continue
            if name == "saved":
                if "saved" in req_data:
                    kwargs['saved'] = req_data['saved']
                else:
                    kwargs['saved'] = sig.parameters['saved'].default
                continue
            if name == "session_id":
                kwargs[name] = request.headers.get("X-Session-Id")
            elif param.annotation in (str, int, dict, list, bool):
                kwargs[name] = req_data.get(name, param.default)
            else:
                kwargs[name] = param.annotation(**req_data)
        return func(*args, **kwargs)

    wrapper.__name__ = func.__name__
    return wrapper


def service_run_queue(service_id, **kwargs):
    """The func used in a separate thread to run an agent service. The result
    will be saved in a queue if one is provided."""
    stream: queue.Queue = kwargs.get('output_stream')
    try:
        res = ServiceInstance(service_id).run(**kwargs)
        return res
    finally:
        if stream:
            stream.put_nowait('{"type": "EOF"}')


def agent_run_queue(agent_id, **kwargs):
    """The func used in a separate thread to run an agent, and the result will be saved in a queue if provided.

    Args:
        agent_id: The agent id
        **kwargs: Arbitrary keyword arguments.
    """
    stream: queue.Queue = kwargs.get('output_stream')
    try:
        agent: Agent = AgentManager().get_instance_obj(agent_id)
        res = agent.run(**kwargs)
        return res
    finally:
        if stream:
            stream.put_nowait('{"type": "EOF"}')


async def async_agent_run_queue(agent_id, **kwargs):
    stream: asyncio.Queue = kwargs.get('output_stream')
    try:
        agent: Agent = await asyncio.to_thread(AgentManager().get_instance_obj, agent_id)
        res = await asyncio.to_thread(agent.run, **kwargs)
        return res
    finally:
        if stream:
            await asyncio.to_thread(stream.put_nowait, '{"type": "EOF"}')


def make_standard_response(success: bool,
                           result=None,
                           message: str = None,
                           request_id: str = None,
                           status_code=200):
    """Construct a standard flask response."""
    response_data = {
        "success": success,
        "result": result,
        "message": message,
        "request_id": request_id
    }
    return make_response(jsonify(response_data), status_code)



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/3/14 16:07
# @Author  : wangchongshi
# @Email   : wangchongshi.wcs@antgroup.com
# @FileName: chat_memory.py
from typing import Optional, List

from langchain.memory.chat_memory import BaseChatMemory

from agentuniverse.agent.memory.enum import MemoryTypeEnum
from agentuniverse.agent.memory.memory import Memory
from agentuniverse.agent.memory.langchain_instance import AuConversationSummaryBufferMemory, \
    AuConversationTokenBufferMemory
from agentuniverse.agent.memory.message import Message
from agentuniverse.base.config.component_configer.configers.memory_configer import MemoryConfiger
from agentuniverse.llm.llm import LLM


class ChatMemory(Memory):
    """The basic class for chat memory model.

    Attributes:
        llm (LLM): the LLM instance used by this memory.
        input_key (Optional[str]): The input key in the model input parameters is used to find the specific query in a
        round of conversations.
        output_key (Optional[str]): The output key in the model output parameters is used to find the specific result
        in a round of conversations.
        prompt_version (Optional[str]): The version of the prompt used for compressing the memory.
        messages (Optional[List[Message]]): The list of conversation messages to send to the LLM memory.
    """

    llm: Optional[LLM] = None
    input_key: Optional[str] = 'input'
    output_key: Optional[str] = 'output'
    prompt_version: Optional[str] = None
    messages: Optional[List[Message]] = None

    def as_langchain(self) -> BaseChatMemory:
        """Convert the agentUniverse(aU) chat memory class to the langchain chat memory class."""
        if self.llm is None:
            raise ValueError("Must set `llm` when using langchain memory.")
        if self.type is None or self.type == MemoryTypeEnum.SHORT_TERM:
            return AuConversationTokenBufferMemory(llm=self.llm.as_langchain(), memory_key=self.memory_key,
                                                   input_key=self.input_key, output_key=self.output_key,
                                                   max_token_limit=self.max_tokens, messages=self.messages)
        elif self.type == MemoryTypeEnum.LONG_TERM:
            return AuConversationSummaryBufferMemory(llm=self.llm.as_langchain(), memory_key=self.memory_key,
                                                     input_key=self.input_key, output_key=self.output_key,
                                                     max_token_limit=self.max_tokens, messages=self.messages,
                                                     prompt_version=self.prompt_version)

    def set_by_agent_model(self, **kwargs):
        """ Assign values of parameters to the ChatMemory model in the agent configuration."""
        copied_obj = super().set_by_agent_model(**kwargs)
        if 'llm' in kwargs and kwargs['llm']:
            copied_obj.llm = kwargs['llm']
        if 'input_key' in kwargs and kwargs['input_key']:
            copied_obj.input_key = kwargs['input_key']
        if 'output_key' in kwargs and kwargs['output_key']:
            copied_obj.output_key = kwargs['output_key']
        return copied_obj

    def initialize_by_component_configer(self, component_configer: MemoryConfiger) -> 'ChatMemory':
        """Initialize the chat memory by the ComponentConfiger object.
        Args:
            component_configer(MemoryConfiger): the ComponentConfiger object
        Returns:
            ChatMemory: the ChatMemory object
        """
        super().initialize_by_component_configer(component_configer)
        if hasattr(component_configer, 'input_key') and component_configer.input_key:
            self.input_key = component_configer.input_key
        if hasattr(component_configer, 'output_key') and component_configer.output_key:
            self.output_key = component_configer.output_key
        if hasattr(component_configer, 'prompt_version') and component_configer.prompt_version:
            self.prompt_version = component_configer.prompt_version
        return self

    def add(self, message_list: List[Message], **kwargs) -> None:
        """Add messages to the memory."""
        if not message_list:
            return
        self.messages = message_list

    def get(self, **kwargs) -> List[Message]:
        """Get messages from the memory."""
        return self.prune(self.messages)



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/7/25 22:50
# @Author  : wangchongshi
# @Email   : wangchongshi.wcs@antgroup.com
# @FileName: agent_product.py
from typing import Optional

from agentuniverse.agent.agent import Agent
from agentuniverse.base.config.application_configer.application_config_manager import ApplicationConfigManager
from agentuniverse_product.base.product import Product
from agentuniverse_product.base.product_configer import ProductConfiger


class AgentProduct(Product):
    """The basic class of the agent product."""

    opening_speech: Optional[str] = None
    _instance: Optional[Agent] = None

    @property
    def instance(self) -> Agent:
        return self._instance

    def get_instance_code(self) -> str:
        """Return the full name of the product."""
        appname = ApplicationConfigManager().app_configer.base_info_appname
        return f"{appname}.product.{self.id}"

    def initialize_by_component_configer(self,
                                         product_configer: ProductConfiger) -> 'Product':
        """Initialize the Product by the ProductConfiger object.

        Args:
            product_configer(ProductConfiger): A configer contains product basic info.
        Returns:
            Product: A Product instance.
        """
        super().initialize_by_component_configer(product_configer)
        if hasattr(product_configer, 'opening_speech') and product_configer.opening_speech:
            self.opening_speech = product_configer.opening_speech
        return self



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/5/20 16:24
# @Author  : wangchongshi
# @Email   : wangchongshi.wcs@antgroup.com
# @FileName: monitor.py
import datetime
import json
import os
import uuid
from typing import Union, Optional

from pydantic import BaseModel

from agentuniverse.agent.input_object import InputObject
from agentuniverse.agent.output_object import OutputObject
from agentuniverse.base.annotation.singleton import singleton
from agentuniverse.base.config.configer import Configer
from agentuniverse.base.context.framework_context_manager import FrameworkContextManager

LLM_INVOCATION_SUBDIR = "llm_invocation"
AGENT_INVOCATION_SUBDIR = "agent_invocation"


@singleton
class Monitor(BaseModel):
    dir: Optional[str] = './monitor'
    activate: Optional[bool] = False

    def __init__(self, configer: Configer = None, **kwargs):
        super().__init__(**kwargs)
        if configer:
            config: dict = configer.value.get('MONITOR', {})
            self.dir = config.get('dir', './monitor')
            self.activate = config.get('activate', False)

    def trace_llm_invocation(self, source: str, llm_input: Union[str, dict], llm_output: Union[str, dict]) -> None:
        """Trace the llm invocation and save it to the monitor jsonl file."""
        if self.activate:
            try:
                import jsonlines
            except ImportError:
                raise ImportError(
                    "jsonlines is required to trace llm invocation: `pip install jsonlines`"
                )
            # get the current time
            date = datetime.datetime.now()
            llm_invocation = {
                "source": source,
                "date": date.strftime("%Y-%m-%d %H:%M:%S"),
                "llm_input": llm_input,
                "llm_output": llm_output,
            }
            # files are stored in hours
            filename = f"llm_{date.strftime('%Y-%m-%d-%H')}.jsonl"
            # file path to save
            path_save = os.path.join(str(self._get_or_create_subdir(LLM_INVOCATION_SUBDIR)), filename)

            # write to jsonl
            with jsonlines.open(path_save, 'a') as writer:
                writer.write(llm_invocation)

    def trace_agent_invocation(self, source: str, agent_input: Union[str, dict],
                               agent_output: Union[str, dict]) -> None:
        """Trace the agent invocation and save it to the monitor jsonl file."""
        if self.activate:
            try:
                import jsonlines
            except ImportError:
                raise ImportError(
                    "jsonlines is required to trace llm invocation: `pip install jsonlines`"
                )
            # get the current time
            date = datetime.datetime.now()
            agent_invocation = {
                "source": source,
                "date": date.strftime("%Y-%m-%d %H:%M:%S"),
                "agent_input": self.serialize_obj(agent_input),
                "agent_output": self.serialize_obj(agent_output),
            }
            # files are stored in hours
            filename = f"agent_{source}_{date.strftime('%Y-%m-%d-%H')}.jsonl"
            # file path to save
            path_save = os.path.join(str(self._get_or_create_subdir(AGENT_INVOCATION_SUBDIR)), filename)

            # write to jsonl
            with jsonlines.open(path_save, 'a') as writer:
                writer.write(agent_invocation)

    @staticmethod
    def init_trace_id():
        """Initialize the trace id in the framework context."""
        if FrameworkContextManager().get_context('trace_id') is None:
            FrameworkContextManager().set_context('trace_id', str(uuid.uuid4()))

    @staticmethod
    def init_invocation_chain():
        """Initialize the invocation chain in the framework context."""
        Monitor.init_trace_id()
        trace_id = FrameworkContextManager().get_context('trace_id')
        if FrameworkContextManager().get_context(trace_id + '_invocation_chain') is None:
            FrameworkContextManager().set_context(trace_id + '_invocation_chain', [])

    @staticmethod
    def clear_invocation_chain():
        """Clear the invocation chain in the framework context."""
        trace_id = FrameworkContextManager().get_context('trace_id')
        if trace_id is not None:
            FrameworkContextManager().del_context(trace_id + '_invocation_chain')

    @staticmethod
    def add_invocation_chain(source: dict):
        """Add the source to the invocation chain"""
        trace_id = FrameworkContextManager().get_context('trace_id')
        if trace_id is not None:
            invocation_chain = FrameworkContextManager().get_context(trace_id + '_invocation_chain')
            if invocation_chain is not None:
                invocation_chain.append(source)
                FrameworkContextManager().set_context(trace_id + '_invocation_chain', invocation_chain)

    @staticmethod
    def get_trace_id():
        """Get the trace id in the framework context."""
        return FrameworkContextManager().get_context('trace_id')

    @staticmethod
    def get_invocation_chain():
        """Get the invocation chain in the framework context."""
        trace_id = FrameworkContextManager().get_context('trace_id')
        return FrameworkContextManager().get_context(trace_id + '_invocation_chain', []) if trace_id is not None else []

    @staticmethod
    def init_token_usage():
        """Initialize the token usage in the framework context."""
        Monitor.init_trace_id()
        trace_id = FrameworkContextManager().get_context('trace_id')
        if FrameworkContextManager().get_context(trace_id + '_token_usage') is None:
            FrameworkContextManager().set_context(trace_id + '_token_usage', {})

    @staticmethod
    def add_token_usage(cur_token_usage: dict):
        """Add the token usage to the framework context."""
        if cur_token_usage is None:
            return
        trace_id = FrameworkContextManager().get_context('trace_id')
        if trace_id is not None:
            old_token_usage: dict = FrameworkContextManager().get_context(trace_id + '_token_usage')
            if old_token_usage is not None:
                for key, value in cur_token_usage.items():
                    old_token_usage[key] = old_token_usage[key] + value if key in old_token_usage else value
                FrameworkContextManager().set_context(trace_id + '_token_usage', old_token_usage)

    @staticmethod
    def clear_token_usage():
        """Clear the token usage in the framework context."""
        trace_id = FrameworkContextManager().get_context('trace_id')
        if trace_id is not None:
            FrameworkContextManager().del_context(trace_id + '_token_usage')
        FrameworkContextManager().del_context('trace_id')

    @staticmethod
    def get_token_usage():
        """Get the token usage in the framework context."""
        trace_id = FrameworkContextManager().get_context('trace_id')
        return FrameworkContextManager().get_context(trace_id + '_token_usage', {}) if trace_id is not None else {}

    def _get_or_create_subdir(self, subdir: str) -> str:
        """Get or create a subdirectory if it doesn't exist in the monitor directory."""
        path = os.path.join(self.dir, subdir)
        os.makedirs(path, exist_ok=True)
        return path

    @staticmethod
    def default_serializer(obj):
        """Default serializer for objects."""
        if isinstance(obj, InputObject):
            return obj.to_dict()
        elif isinstance(obj, OutputObject):
            return obj.to_dict()
        elif isinstance(obj, BaseModel):
            try:
                return obj.dict()
            except TypeError:
                raise TypeError(f"Object of type {type(obj).__name__} is not JSON serializable")
        else:
            raise TypeError(f"Object of type {type(obj).__name__} is not JSON serializable")

    def serialize_obj(self, obj):
        """Serialize an object and filter out non-serializable values."""
        filtered_obj = self.filter_and_serialize(obj)
        return json.loads(json.dumps(filtered_obj, default=self.default_serializer))

    def filter_and_serialize(self, obj):
        """Recursively filter out non-serializable values from an object."""

        def is_json_serializable(value):
            """Check if value is a JSON serializable object."""
            try:
                json.dumps(value, default=self.default_serializer)
                return True
            except (TypeError, OverflowError):
                return False

        def filter_dict(d):
            return {k: v for k, v in d.items() if is_json_serializable(v)}

        def recursive_filter(o):
            if isinstance(o, dict):
                return filter_dict({k: recursive_filter(v) for k, v in o.items()})
            elif isinstance(o, list):
                return [recursive_filter(i) for i in o if is_json_serializable(i)]
            else:
                return o

        return recursive_filter(obj)



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/4/29 16:00
# @Author  : wangchongshi
# @Email   : wangchongshi.wcs@antgroup.com
# @FileName: peer_agent.py
from agentuniverse.agent.template.peer_agent_template import PeerAgentTemplate


class PeerAgent(PeerAgentTemplate):
    """Peer Agent class."""



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/10/17 21:08
# @Author  : wangchongshi
# @Email   : wangchongshi.wcs@antgroup.com
# @FileName: expressing_agent.py
from agentuniverse.agent.template.expressing_agent_template import ExpressingAgentTemplate


class ExpressingAgent(ExpressingAgentTemplate):
    """Expressing Agent class."""



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/8/29 11:13
# @Author  : fanen.lhy
# @Email   : fanen.lhy@antgroup.com
# @FileName: summary_agent.py

from agentuniverse.agent.agent import Agent
from agentuniverse.agent.input_object import InputObject


class SimpleSummaryAgent(Agent):
    """ReAct Agent class."""

    def input_keys(self) -> list[str]:
        """Return the input keys of the Agent."""
        return ['input']

    def output_keys(self) -> list[str]:
        """Return the output keys of the Agent."""
        return ['output']

    def parse_input(self, input_object: InputObject, agent_input: dict) -> dict:
        """Agent parameter parsing.

        Args:
            input_object (InputObject): input parameters passed by the user.
            agent_input (dict): agent input preparsed by the agent.
        Returns:
            dict: agent input parsed from `input_object` by the user.
        """
        agent_input['input'] = input_object.get_data('input')
        self.agent_model.profile.setdefault('prompt_version', 'simple_summary_agent.cn')
        return agent_input

    def parse_result(self, planner_result: dict) -> dict:
        """Planner result parser.

        Args:
            planner_result(dict): Planner result
        Returns:
            dict: Agent result object.
        """
        return planner_result


# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/10/10 18:53
# @Author  : wangchongshi
# @Email   : wangchongshi.wcs@antgroup.com
# @FileName: memory_storage_manager.py
from agentuniverse.agent.memory.memory_storage.memory_storage import MemoryStorage
from agentuniverse.base.annotation.singleton import singleton
from agentuniverse.base.component.component_enum import ComponentEnum
from agentuniverse.base.component.component_manager_base import ComponentManagerBase


@singleton
class MemoryStorageManager(ComponentManagerBase[MemoryStorage]):
    """A singleton manager class of the MemoryStorage."""

    def __init__(self):
        super().__init__(ComponentEnum.MEMORY_STORAGE)



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/8/23 15:35
# @Author  : wangchongshi
# @Email   : wangchongshi.wcs@antgroup.com
# @FileName: workflow_planner.py
from agentuniverse.agent.agent_model import AgentModel
from agentuniverse.agent.input_object import InputObject
from agentuniverse.agent.plan.planner.planner import Planner
from agentuniverse.workflow.workflow import Workflow
from agentuniverse.workflow.workflow_manager import WorkflowManager
from agentuniverse.workflow.workflow_output import WorkflowOutput


class WorkflowPlanner(Planner):
    """Workflow planner class."""

    def invoke(self, agent_model: AgentModel, planner_input: dict, input_object: InputObject) -> dict:
        """Invoke the planner.

        Args:
            agent_model (AgentModel): Agent model object.
            planner_input (dict): Planner input object.
            input_object (InputObject): The input parameters passed by the user.
        Returns:
            dict: The planner result.
        """
        planner = agent_model.plan.get('planner', {})
        workflow_id = planner.get('workflow_id')
        workflow: Workflow = WorkflowManager().get_instance_obj(component_instance_name=workflow_id)
        # build and run workflow
        if workflow.graph_config is None:
            raise Exception('Workflow graph is None, please add nodes and edges to the workflow graph.')
        workflow = workflow.build()
        workflow_output: WorkflowOutput = workflow.run(input_object.to_dict())
        print(workflow_output.workflow_node_results)
        return workflow_output.workflow_end_params



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/8/20 20:03
# @Author  : wangchongshi
# @Email   : wangchongshi.wcs@antgroup.com
# @FileName: node_output.py
from typing import Optional, Dict, Any

from pydantic import BaseModel

from agentuniverse.workflow.node.enum import NodeStatusEnum


class NodeOutput(BaseModel):
    """The basic class of the node output."""

    node_id: Optional[str] = None
    result: Optional[Any] = None
    error: Optional[str] = None
    status: NodeStatusEnum = NodeStatusEnum.RUNNING
    metadata: Optional[Dict[str, Any]] = None
    edge_source_handler: Optional[str] = None



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/4/2 16:15
# @Author  : wangchongshi
# @Email   : wangchongshi.wcs@antgroup.com
# @FileName: openai_llm.py
from typing import Any, Optional, AsyncIterator, Iterator, Union

import httpx
from langchain_core.language_models.base import BaseLanguageModel
from openai import OpenAI, AsyncOpenAI
from pydantic import Field
import tiktoken

from agentuniverse.llm.langchain_instance import LangchainOpenAI
from agentuniverse.llm.llm import LLM, LLMOutput
from agentuniverse.base.util.env_util import get_from_env

OPENAI_MAX_CONTEXT_LENGTH = {
    "gpt-3.5-turbo": 4096,
    "gpt-3.5-turbo-0301": 4096,
    "gpt-3.5-turbo-0613": 4096,
    "gpt-3.5-turbo-16k": 16384,
    "gpt-3.5-turbo-16k-0613": 16384,
    "gpt-35-turbo": 4096,
    "gpt-35-turbo-16k": 16384,
    "gpt-3.5-turbo-1106": 16384,
    "gpt-3.5-turbo-0125": 16384,
    "gpt-4-0314": 8192,
    "gpt-4": 8192,
    "gpt-4-32k": 32768,
    "gpt-4-32k-0613": 32768,
    "gpt-4-0613": 8192,
    "gpt-4-1106-preview": 128000,
    "gpt-4-turbo": 128000,
    "gpt-4o": 128000,
    "gpt-4o-2024-05-13": 128000,
}


class OpenAILLM(LLM):
    """The openai llm class.

    Attributes:
        openai_api_key (Optional[str], optional): The API key for the OpenAI API.
        This automatically infers the `openai_api_key` from the environment variable `OPENAI_API_KEY` if not provided.

        openai_organization (Optional[str], optional): The OpenAI organization.
        This automatically infers the `openai_organization` from the environment variable `OPENAI_ORGANIZATION` if not provided.

        openai_api_base (Optional[str], optional): The OpenAI base url.
        This automatically infers the `openai_api_base` from the environment variable `OPENAI_API_BASE` if not provided.

        openai_client_args (Optional[dict], optional): Additional arguments to pass to the OpenAI client.
   """

    openai_api_key: Optional[str] = Field(default_factory=lambda: get_from_env("OPENAI_API_KEY"))
    openai_organization: Optional[str] = Field(default_factory=lambda: get_from_env("OPENAI_ORGANIZATION"))
    openai_api_base: Optional[str] = Field(default_factory=lambda: get_from_env("OPENAI_API_BASE"))
    openai_proxy: Optional[str] = Field(default_factory=lambda: get_from_env("OPENAI_PROXY"))
    openai_client_args: Optional[dict] = None

    def _new_client(self):
        """Initialize the openai client."""
        return OpenAI(
            api_key=self.openai_api_key,
            organization=self.openai_organization,
            base_url=self.openai_api_base,
            timeout=self.request_timeout,
            max_retries=self.max_retries,
            http_client=httpx.Client(proxy=self.openai_proxy) if self.openai_proxy else None,
            **(self.openai_client_args or {}),
        )

    def _new_async_client(self):
        """Initialize the openai async client."""
        return AsyncOpenAI(
            api_key=self.openai_api_key,
            organization=self.openai_organization,
            base_url=self.openai_api_base,
            timeout=self.request_timeout,
            max_retries=self.max_retries,
            http_client=httpx.AsyncClient(proxy=self.openai_proxy) if self.openai_proxy else None,
            **(self.openai_client_args or {}),
        )

    def _call(self, messages: list, **kwargs: Any) -> Union[LLMOutput, Iterator[LLMOutput]]:
        """Run the OpenAI LLM.

        Args:
            messages (list): The messages to send to the LLM.
            **kwargs: Arbitrary keyword arguments.
        """
        streaming = kwargs.pop("streaming") if "streaming" in kwargs else self.streaming
        self.client = self._new_client()
        chat_completion = self.client.chat.completions.create(
            messages=messages,
            model=kwargs.pop('model', self.model_name),
            temperature=kwargs.pop('temperature', self.temperature),
            stream=kwargs.pop('stream', streaming),
            max_tokens=kwargs.pop('max_tokens', self.max_tokens),
            **kwargs,
        )
        if not streaming:
            text = chat_completion.choices[0].message.content
            return LLMOutput(text=text, raw=chat_completion.model_dump())
        return self.generate_stream_result(chat_completion)

    async def _acall(self, messages: list, **kwargs: Any) -> Union[LLMOutput, AsyncIterator[LLMOutput]]:
        """Asynchronously run the OpenAI LLM.

        Args:
            messages (list): The messages to send to the LLM.
            **kwargs: Arbitrary keyword arguments.
        """
        streaming = kwargs.pop("streaming") if "streaming" in kwargs else self.streaming
        self.async_client = self._new_async_client()
        chat_completion = await self.async_client.chat.completions.create(
            messages=messages,
            model=kwargs.pop('model', self.model_name),
            temperature=kwargs.pop('temperature', self.temperature),
            stream=kwargs.pop('stream', streaming),
            max_tokens=kwargs.pop('max_tokens', self.max_tokens),
            **kwargs,
        )
        if not streaming:
            text = chat_completion.choices[0].message.content
            return LLMOutput(text=text, raw=chat_completion.model_dump())
        return self.agenerate_stream_result(chat_completion)

    def as_langchain(self) -> BaseLanguageModel:
        """Convert the agentUniverse(aU) openai llm class to the langchain openai llm class."""
        return LangchainOpenAI(self)

    def set_by_agent_model(self, **kwargs):
        """ Assign values of parameters to the OpenAILLM model in the agent configuration."""
        copied_obj = super().set_by_agent_model(**kwargs)
        if 'openai_api_key' in kwargs and kwargs['openai_api_key']:
            copied_obj.openai_api_key = kwargs['openai_api_key']
        if 'openai_api_base' in kwargs and kwargs['openai_api_base']:
            copied_obj.openai_api_base = kwargs['openai_api_base']
        if 'openai_proxy' in kwargs and kwargs['openai_proxy']:
            copied_obj.openai_proxy = kwargs['openai_proxy']
        if 'openai_client_args' in kwargs and kwargs['openai_client_args']:
            copied_obj.openai_client_args = kwargs['openai_client_args']
        return copied_obj

    def max_context_length(self) -> int:
        """Max context length.

          The total length of input tokens and generated tokens is limited by the openai model's context length.
          """
        return OPENAI_MAX_CONTEXT_LENGTH.get(self.model_name, 4096)

    def get_num_tokens(self, text: str) -> int:
        """Get the number of tokens present in the text.

        Useful for checking if an input will fit in an openai model's context window.

        Args:
            text: The string input to tokenize.

        Returns:
            The integer number of tokens in the text.
        """
        try:
            encoding = tiktoken.encoding_for_model(self.model_name)
        except KeyError:
            encoding = tiktoken.get_encoding("cl100k_base")
        return len(encoding.encode(text))

    @staticmethod
    def parse_result(chunk):
        """Generate the result of the stream."""
        chat_completion = chunk
        if not isinstance(chunk, dict):
            chunk = chunk.dict()
        if len(chunk["choices"]) == 0:
            return
        choice = chunk["choices"][0]
        message = choice.get("delta")
        text = message.get("content")
        if not text:
            return
        return LLMOutput(text=text, raw=chat_completion.model_dump())

    @classmethod
    def generate_stream_result(cls, stream: Iterator) -> Iterator[LLMOutput]:
        """Generate the result of the stream."""
        for chunk in stream:
            llm_output = cls.parse_result(chunk)
            if llm_output:
                yield llm_output

    @classmethod
    async def agenerate_stream_result(cls, stream: AsyncIterator) -> AsyncIterator[LLMOutput]:
        """Generate the result of the stream."""
        async for chunk in stream:
            llm_output = cls.parse_result(chunk)
            if llm_output:
                yield llm_output



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/10/8 10:48
# @Author  : wangchongshi
# @Email   : wangchongshi.wcs@antgroup.com
# @FileName: work_pattern_configer.py
from typing import Optional

from agentuniverse.base.config.component_configer.component_configer import ComponentConfiger
from agentuniverse.base.config.configer import Configer


class WorkPatternConfiger(ComponentConfiger):
    """The WorkPatternConfiger class, which is used to load and manage the WorkPattern configuration."""

    def __init__(self, configer: Optional[Configer] = None):
        """Initialize the WorkPatternConfiger."""
        super().__init__(configer)
        self.__name: Optional[str] = None
        self.__description: Optional[str] = None

    @property
    def name(self) -> Optional[str]:
        """Return the name of the WorkPattern."""
        return self.__name

    @property
    def description(self) -> Optional[str]:
        """Return the description of the WorkPattern."""
        return self.__description

    def load(self) -> 'WorkPatternConfiger':
        """Load the configuration by the Configer object.
        Returns:
            WorkPatternConfiger: the WorkPatternConfiger object
        """
        return self.load_by_configer(self.__configer)

    def load_by_configer(self, configer: Configer) -> 'WorkPatternConfiger':
        """Load the configuration by the Configer object.
        Args:
            configer(Configer): the Configer object
        Returns:
            WorkPatternConfiger: the WorkPatternConfiger object
        """
        super().load_by_configer(configer)
        try:
            self.__name = configer.value.get('name')
            self.__description = configer.value.get('description')
        except Exception as e:
            raise Exception(f"Failed to parse the work pattern configuration: {e}")
        return self



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/8/21 18:46
# @Author  : wangchongshi
# @Email   : wangchongshi.wcs@antgroup.com
# @FileName: end_node.py
import re
from typing import List, Optional

from agentuniverse.workflow.node.enum import NodeEnum, NodeStatusEnum
from agentuniverse.workflow.node.node import Node, NodeData
from agentuniverse.workflow.node.node_config import EndNodeInputParams, NodeOutputParams, NodeInfoParams
from agentuniverse.workflow.node.node_output import NodeOutput
from agentuniverse.workflow.workflow_output import WorkflowOutput


class EndNodeData(NodeData):
    inputs: Optional[EndNodeInputParams] = None


class EndNode(Node):
    """The basic class of the end node."""
    _data_cls = EndNodeData

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.type = NodeEnum.END

    def _run(self, workflow_output: WorkflowOutput) -> NodeOutput:
        inputs: EndNodeInputParams = self._data.inputs
        prompt: NodeInfoParams = inputs.prompt
        if isinstance(prompt.value, dict):
            prompt_val = prompt.value.get('content', '')
        else:
            prompt_val = prompt.value

        # Extract variables from the prompt template
        template_variables = re.findall(r'\{\{(.*?)\}\}', prompt_val)
        # Resolve the input parameters
        end_node_input_params = self._resolve_input_params(inputs.input_param, workflow_output)

        # Replace variables in the prompt
        try:
            for var in template_variables:
                if var not in end_node_input_params:
                    raise KeyError(f"The variable '{var}' is not found in the input params.")
                prompt_val = prompt_val.replace(f'{{{{{var}}}}}',
                                                str(end_node_input_params[var]) if end_node_input_params[var] else '')
        except KeyError as e:
            raise ValueError(f"Error processing template variables: {e}")

        output_params: List[NodeOutputParams] = self._data.outputs
        output_param: NodeOutputParams = output_params[0]
        output_param.value = prompt_val

        workflow_output.workflow_parameters[self.id] = output_params
        workflow_output.workflow_end_params = {output_param.name: output_param.value}
        return NodeOutput(node_id=self.id, status=NodeStatusEnum.SUCCEEDED, result=output_params)



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/4/2 15:21
# @Author  : jerry.zzw 
# @Email   : jerry.zzw@antgroup.com
# @FileName: singleton.py
from functools import wraps


def singleton(cls):
    """Decorator to make a class a Singleton class (only one instance), using closure."""
    instances = {}

    @wraps(cls)
    def get_instance(*args, **kwargs):
        if cls not in instances:
            instances[cls] = cls(*args, **kwargs)
        return instances[cls]

    return get_instance



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/4/16 14:42
# @Author  : wangchongshi
# @Email   : wangchongshi.wcs@antgroup.com
# @FileName: prompt_util.py
from typing import List

from langchain.chains.summarize import load_summarize_chain
from langchain_core.documents import Document

from agentuniverse.agent.memory.enum import ChatMessageEnum
from agentuniverse.agent.memory.message import Message
from agentuniverse.llm.llm import LLM
from agentuniverse.llm.llm_manager import LLMManager
from agentuniverse.prompt.prompt_manager import PromptManager
from agentuniverse.prompt.prompt_model import AgentPromptModel
from agentuniverse.prompt.enum import PromptProcessEnum


def summarize_by_stuff(texts: List[str], llm: LLM, summary_prompt):
    """
    stuff summarization -- general method
    """
    stuff_chain = load_summarize_chain(llm.as_langchain(), chain_type='stuff', verbose=True,
                                       prompt=summary_prompt.as_langchain())
    return stuff_chain.run([Document(page_content=text) for text in texts])


def summarize_by_map_reduce(texts: List[str], llm: LLM, summary_prompt, combine_prompt):
    """
    map reduce summarization -- general method
    """
    map_reduce_chain = load_summarize_chain(llm.as_langchain(), chain_type='map_reduce', verbose=True,
                                            map_prompt=summary_prompt.as_langchain(),
                                            combine_prompt=combine_prompt.as_langchain())
    return map_reduce_chain.run([Document(page_content=text) for text in texts])


def split_text_on_tokens(text: str, text_token: int, chunk_size=800, chunk_overlap=100) -> List[str]:
    """Split incoming text and return chunks using tokenizer."""
    # calculate the number of characters represented by each token.
    char_per_token = len(text) / text_token
    chunk_char_size = int(chunk_size * char_per_token)
    chunk_char_overlap = int(chunk_overlap * char_per_token)

    result = []
    current_position = 0

    while current_position + chunk_char_overlap < len(text):
        if current_position + chunk_char_size >= len(text):
            chunk = text[current_position:]
        else:
            chunk = text[current_position:current_position + chunk_char_size]

        result.append(chunk)
        current_position += chunk_char_size - chunk_char_overlap

    if len(result) == 0:
        result.append(text[current_position:])

    return result


def split_texts(texts: list[str], llm: LLM, chunk_size=800, chunk_overlap=100, retry=True) -> list[str]:
    """
    split texts into chunks with the fixed token length -- general method
    """
    try:
        split_texts_res = []
        for text in texts:
            text_token = llm.get_num_tokens(text)
            split_texts_res.extend(
                split_text_on_tokens(text=text, text_token=text_token, chunk_size=chunk_size,
                                     chunk_overlap=chunk_overlap))
        return split_texts_res
    except Exception as e:
        if retry:
            return split_texts(texts=texts, llm=llm, retry=False)
        raise ValueError("split text failed, exception=" + str(e))


def truncate_content(content: str, token_length: int, llm: LLM) -> str:
    """
    truncate the content based on the llm token limit
    """
    return str(split_texts(texts=[content], chunk_size=token_length, llm=llm)[0])


def generate_template(agent_prompt_model: AgentPromptModel, prompt_assemble_order: list[str]) -> str:
    """Convert the agent prompt model to an ordered list.

    Args:
        agent_prompt_model (AgentPromptModel): The agent prompt model.
        prompt_assemble_order (list[str]): The prompt assemble ordered list.
    Returns:
        list: The ordered list.
    """
    values = []
    for attr in prompt_assemble_order:
        value = getattr(agent_prompt_model, attr, None)
        if value is not None:
            values.append(value)

    return "\n".join(values)


def generate_chat_template(agent_prompt_model: AgentPromptModel, prompt_assemble_order: list[str]) -> list[Message]:
    """Convert the agent prompt model to the agentUniverse message list.

    Args:
        agent_prompt_model (AgentPromptModel): The agent prompt model.
        prompt_assemble_order (list[str]): The prompt assemble ordered list.
    Returns:
        list: The agentUniverse message list.
    """
    message_list = []
    for attr in prompt_assemble_order:
        value = getattr(agent_prompt_model, attr, None)
        if value is not None:
            message_list.append(
                Message(type=agent_prompt_model.get_message_type(attr), content=value))
    if message_list:
        # Integrate the system messages and put them in the first of the message list.
        system_messages = '\n'.join(msg.content for msg in message_list if msg.type == ChatMessageEnum.SYSTEM.value)
        if system_messages:
            message_list = list(filter(lambda msg: msg.type != ChatMessageEnum.SYSTEM.value, message_list))
            message_list.insert(0, Message(type=ChatMessageEnum.SYSTEM.value, content=system_messages))
    return message_list


def process_llm_token(agent_llm: LLM, lc_prompt_template, profile: dict, planner_input: dict):
    """Process the prompt template based on the prompt processor.

    Args:
        agent_llm (LLM): The agent llm.
        lc_prompt_template: The langchain prompt template.
        profile (dict): The profile.
        planner_input (dict): The planner input.
    """
    llm_model: dict = profile.get('llm_model')

    # get the prompt processor configuration
    prompt_processor: dict = llm_model.get('prompt_processor') or dict()
    prompt_processor_type: str = prompt_processor.get('type') or PromptProcessEnum.TRUNCATE.value
    prompt_processor_llm: str = prompt_processor.get('llm')

    # get the summary and combine prompt versions
    summary_prompt_version: str = prompt_processor.get('summary_prompt_version') or 'prompt_processor.summary_cn'
    combine_prompt_version: str = prompt_processor.get('combine_prompt_version') or 'prompt_processor.combine_cn'

    prompt_input_dict = {key: planner_input[key] for key in lc_prompt_template.input_variables if key in planner_input}

    # get the llm instance for prompt compression
    prompt_llm: LLM = LLMManager().get_instance_obj(prompt_processor_llm)

    if prompt_llm is None:
        prompt_llm = agent_llm

    prompt = lc_prompt_template.format(**prompt_input_dict)
    # get the number of tokens in the prompt
    prompt_tokens: int = agent_llm.get_num_tokens(prompt)

    input_tokens = agent_llm.max_context_length() - agent_llm.max_tokens
    if input_tokens <= 0:
        raise Exception("The current output max tokens limit is greater than the context length of the LLM model, "
                        "please adjust it by editing the `max_tokens` parameter in the llm yaml.")

    if prompt_tokens <= input_tokens:
        return

    process_prompt_type_enum = PromptProcessEnum.from_value(prompt_processor_type)

    # compress the background in the prompt
    content = planner_input.get('background')

    if content:
        if process_prompt_type_enum == PromptProcessEnum.TRUNCATE:
            planner_input['background'] = truncate_content(content, input_tokens, agent_llm)
        elif process_prompt_type_enum == PromptProcessEnum.STUFF:
            planner_input['background'] = summarize_by_stuff(texts=[content], llm=prompt_llm,
                                                             summary_prompt=PromptManager().get_instance_obj(
                                                                 summary_prompt_version))
        elif process_prompt_type_enum == PromptProcessEnum.MAP_REDUCE:
            planner_input['background'] = summarize_by_map_reduce(texts=split_texts([content], agent_llm),
                                                                  llm=prompt_llm,
                                                                  summary_prompt=PromptManager().get_instance_obj(
                                                                      summary_prompt_version),
                                                                  combine_prompt=PromptManager().get_instance_obj(
                                                                      combine_prompt_version))



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/8/27 23:16
# @Author  : wangchongshi
# @Email   : wangchongshi.wcs@antgroup.com
# @FileName: plugin_util.py
import re
from typing import Dict, List
import uuid

from yaml import safe_load
from agentuniverse_product.base.product_manager import ProductManager
from agentuniverse_product.service.model.plugin_dto import PluginDTO


def validate_create_plugin_parameters(plugin_dto: PluginDTO) -> None:
    """Validate the parameters for creating a plugin instance.

    Args:
        plugin_dto (PluginDTO): The plugin DTO object containing the plugin parameters.
    """
    if plugin_dto.id is None:
        raise ValueError("Plugin id cannot be None.")
    plugin = ProductManager().get_instance_obj(plugin_dto.id)
    if plugin:
        raise ValueError("Plugin instance corresponding to the plugin id already exists.")
    if plugin_dto.openapi_desc is None:
        raise ValueError("The openapi_desc in plugin cannot be None.")


def assemble_plugin_product_config_data(plugin_dto: PluginDTO, tool_id_list: List[str]) -> Dict:
    """Assemble the plugin product configuration data.

    Args:
        plugin_dto (PluginDTO): The plugin DTO object containing the plugin parameters.
        tool_id_list (List[str]): The list of tool IDs associated with the plugin.

    Returns:
        Dict: The assembled plugin product configuration data.
    """
    return {
        'id': plugin_dto.id,
        'nickname': plugin_dto.nickname,
        'avatar': plugin_dto.avatar,
        'description': plugin_dto.description,
        'type': 'PLUGIN',
        'toolset': tool_id_list,
        'metadata': {
            'class': 'PluginProduct',
            'module': 'agentuniverse_product.base.plugin_product',
            'type': 'PRODUCT'
        },
        'openapi_desc': ''
    }


def parse_openapi_yaml_to_tool_bundle(yaml: str) -> list:
    """Parse the openapi schema yaml to tool bundles.

    Args:
        yaml (str): The openapi schema yaml string.

    Returns:
        list: The list of tool bundles.
    """

    openapi: dict = safe_load(yaml)
    if openapi is None:
        raise Exception('Invalid openapi yaml.')

    if len(openapi['servers']) == 0:
        raise Exception('No server found in the openapi yaml.')

    server_url = str(openapi['servers'][0]['url'])

    # list all interfaces
    interfaces = []
    for path, path_item in openapi['paths'].items():
        methods = ['get', 'post', 'put', 'delete',
                   'patch', 'head', 'options', 'trace']
        for method in methods:
            if method in path_item:
                interfaces.append({
                    'path': path,
                    'method': method,
                    'operation': path_item[method],
                    'url': server_url + path,
                })
    # create tool bundle
    for interface in interfaces:
        # check if there is a request body
        if 'requestBody' in interface['operation']:
            request_body = interface['operation']['requestBody']
            if 'content' in request_body:
                for content_type, content in request_body['content'].items():
                    # if there is a reference, get the reference and overwrite the content
                    if 'schema' not in content:
                        continue

                    if '$ref' in content['schema']:
                        # get the reference
                        root = openapi
                        reference = content['schema']['$ref'].split(
                            '/')[1:]
                        for ref in reference:
                            root = root[ref]
                        # overwrite the content
                        interface['operation']['requestBody']['content'][content_type]['schema'] = root

        if 'operationId' not in interface['operation']:
            # remove special characters like / to ensure the operation id is valid ^[a-zA-Z0-9_-]{1,64}$
            path = interface['path']
            if interface['path'].startswith('/'):
                path = interface['path'][1:]
            path = re.sub(r'[^a-zA-Z0-9_-]', '', path)
            if not path:
                path = str(uuid.uuid4())

            interface['operation']['operationId'] = f'{path}_{interface["method"]}'

    return interfaces



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

from agentuniverse.agent_serve.web.rpc.grpc import \
    agentuniverse_service_pb2 as agentuniverse__service__pb2


class AgentUniverseServiceStub(object):
    """RPC 服务定义
    """

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.service_run = channel.unary_unary(
                '/agentuniverse.AgentUniverseService/service_run',
                request_serializer=agentuniverse__service__pb2.AgentServiceRequest.SerializeToString,
                response_deserializer=agentuniverse__service__pb2.AgentServiceResponse.FromString,
                )
        self.service_run_async = channel.unary_unary(
                '/agentuniverse.AgentUniverseService/service_run_async',
                request_serializer=agentuniverse__service__pb2.AgentServiceRequest.SerializeToString,
                response_deserializer=agentuniverse__service__pb2.AgentServiceResponse.FromString,
                )
        self.service_run_result = channel.unary_unary(
                '/agentuniverse.AgentUniverseService/service_run_result',
                request_serializer=agentuniverse__service__pb2.AgentResultRequest.SerializeToString,
                response_deserializer=agentuniverse__service__pb2.AgentServiceResponse.FromString,
                )


class AgentUniverseServiceServicer(object):
    """RPC 服务定义
    """

    def service_run(self, request, context):
        """方法定义
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def service_run_async(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def service_run_result(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_AgentUniverseServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'service_run': grpc.unary_unary_rpc_method_handler(
                    servicer.service_run,
                    request_deserializer=agentuniverse__service__pb2.AgentServiceRequest.FromString,
                    response_serializer=agentuniverse__service__pb2.AgentServiceResponse.SerializeToString,
            ),
            'service_run_async': grpc.unary_unary_rpc_method_handler(
                    servicer.service_run_async,
                    request_deserializer=agentuniverse__service__pb2.AgentServiceRequest.FromString,
                    response_serializer=agentuniverse__service__pb2.AgentServiceResponse.SerializeToString,
            ),
            'service_run_result': grpc.unary_unary_rpc_method_handler(
                    servicer.service_run_result,
                    request_deserializer=agentuniverse__service__pb2.AgentResultRequest.FromString,
                    response_serializer=agentuniverse__service__pb2.AgentServiceResponse.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'agentuniverse.AgentUniverseService', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class AgentUniverseService(object):
    """RPC 服务定义
    """

    @staticmethod
    def service_run(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/agentuniverse.AgentUniverseService/service_run',
            agentuniverse__service__pb2.AgentServiceRequest.SerializeToString,
            agentuniverse__service__pb2.AgentServiceResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def service_run_async(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/agentuniverse.AgentUniverseService/service_run_async',
            agentuniverse__service__pb2.AgentServiceRequest.SerializeToString,
            agentuniverse__service__pb2.AgentServiceResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def service_run_result(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/agentuniverse.AgentUniverseService/service_run_result',
            agentuniverse__service__pb2.AgentResultRequest.SerializeToString,
            agentuniverse__service__pb2.AgentServiceResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/3/18 17:34
# @Author  : wangchongshi
# @Email   : wangchongshi.wcs@antgroup.com
# @FileName: pptx_parser.py
from pathlib import Path
from typing import Dict, List, Optional, Union

from agentuniverse.agent.action.knowledge.reader.reader import Reader
from agentuniverse.agent.action.knowledge.store.document import Document


class PptxReader(Reader):
    """Pptx reader."""

    def _load_data(self, file: Union[str, Path], ext_info: Optional[Dict] = None) -> List[Document]:
        """Parse the pptx file.

        Note:
            `python-pptx` is required to read PPTX files: `pip install python-pptx`
        """
        try:
            from pptx import Presentation
        except ImportError:
            raise ImportError(
                "python-pptx is required to read pptx files: `pip install python-pptx`"
            )
        if isinstance(file, str):
            file = Path(file)
        presentation = Presentation(file)
        document_list = []
        for slide_number, slide in enumerate(presentation.slides, start=1):
            for shape in slide.shapes:
                if hasattr(shape, "text"):
                    metadata = {"slide_number": slide_number, "file_name": file.name}
                    if ext_info is not None:
                        metadata.update(ext_info)
                    # Extract the text from the shape
                    document_list.append(Document(text=shape.text, metadata=metadata))
        return document_list



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/7/25 23:27
# @Author  : wangchongshi
# @Email   : wangchongshi.wcs@antgroup.com
# @FileName: llm_dto.py
from typing import Optional, List

from pydantic import BaseModel, Field, ConfigDict


class LlmDTO(BaseModel):
    id: str = Field(description="ID")
    nickname: Optional[str] = Field(description="llm nickname", default="")
    temperature: Optional[float] = Field(description="llm temperature", default=None)
    model_name: Optional[List[str]] = Field(description="llm model name list", default=[])

    # pydantic protected_namespaces config
    model_config = ConfigDict(protected_namespaces=())



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/6/13 11:03
# @Author  : weizjajj 
# @Email   : weizhongjie.wzj@antgroup.com
# @FileName: stream_callback.py

import asyncio
from typing import Optional, Dict, Any, Union
from uuid import UUID

from langchain_core.agents import AgentAction, AgentFinish
from langchain_core.callbacks import BaseCallbackHandler
from langchain_core.outputs import GenerationChunk, ChatGenerationChunk


class StreamOutPutCallbackHandler(BaseCallbackHandler):
    """Callback Handler that prints to std out."""

    def __init__(self, queue_stream: asyncio.Queue, color: Optional[str] = None, agent_info: dict = None) -> None:
        """Initialize callback handler."""
        self.queueStream = queue_stream
        self.color = color
        if agent_info is None:
            agent_info = {}
        self.agent_info = agent_info

    def on_chain_start(
            self, serialized: Dict[str, Any], inputs: Dict[str, Any], **kwargs: Any
    ) -> None:
        return

    def on_chain_end(self, outputs: Dict[str, Any], **kwargs: Any) -> None:
        """Print out that we finished a chain."""

    def on_agent_action(
            self, action: AgentAction, color: Optional[str] = None, **kwargs: Any
    ) -> Any:
        self.queueStream.put_nowait({
            "type": "ReAct",
            "data": {
                "output": "\nThought:" + action.log,
                "agent_info": self.agent_info
            }
        })

    def on_llm_new_token(
            self,
            token: str,
            *,
            chunk: Optional[Union[GenerationChunk, ChatGenerationChunk]] = None,
            run_id: UUID,
            parent_run_id: Optional[UUID] = None,
            **kwargs: Any,
    ) -> Any:
        # add token chunk to the queue.
        self.queueStream.put_nowait({
            "type": "token",
            "data": {
                "chunk": chunk.text,
                "agent_info": self.agent_info
            }
        })

    def on_tool_end(
            self,
            output: str,
            color: Optional[str] = None,
            observation_prefix: Optional[str] = None,
            llm_prefix: Optional[str] = None,
            **kwargs: Any,
    ) -> None:
        """If not the final action, print out observation."""
        if observation_prefix is not None:
            self.queueStream.put_nowait({
                "type": "ReAct",
                "data": {
                    "output": '\n' + observation_prefix + output,
                    "agent_info": self.agent_info
                }
            })
        else:
            self.queueStream.put_nowait({
                "type": "ReAct",
                "data": {
                    "output": '\n Observation:' + output,
                    "agent_info": self.agent_info
                }
            })

    def on_text(
            self,
            text: str,
            color: Optional[str] = None,
            end: str = "",
            **kwargs: Any,
    ) -> None:
        """Run when agent ends."""

    def on_agent_finish(
            self, finish: AgentFinish, color: Optional[str] = None, **kwargs: Any
    ) -> None:
        """Run on agent end."""
        self.queueStream.put_nowait({
            "type": "ReAct",
            "data": {
                "output": '\nThought:' + finish.output,
                "agent_info": self.agent_info
            }
        })



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/7/24 12:00
# @Author  : fanen.lhy
# @Email   : fanen.lhy@antgroup.com
# @FileName: rag_router.py

from abc import abstractmethod
from typing import List, Optional, Tuple

from agentuniverse.agent.action.knowledge.store.query import Query
from agentuniverse.base.component.component_base import ComponentEnum
from agentuniverse.base.component.component_base import ComponentBase
from agentuniverse.base.config.component_configer.component_configer import ComponentConfiger


class RagRouter(ComponentBase):
    """The basic class for rag router."""

    component_type: ComponentEnum = ComponentEnum.RAG_ROUTER
    name: Optional[str] = None
    description: Optional[str] = None

    def rag_route(self, query: Query, store_list: List[str]) \
            -> List[Tuple[Query, str]]:
        """Accept query a list of store instance name, and return a list of
         query-store pair."""
        return self._rag_route(query, store_list)

    def _rag_route(self, query: Query, store_list: List[str]) \
            -> List[Tuple[Query, str]]:
        """Accept query a list of store instance name, and return a list of
         query-store pair."""
        pass

    def _initialize_by_component_configer(self,
                                         rag_router_config: ComponentConfiger) \
            -> 'RagRouter':
        """Initialize the rag router by the ComponentConfiger object.

        Args:
            rag_router_config(ComponentConfiger): A configer contains rag router
            basic info.
        Returns:
            RagRouter: A rag router instance.
        """
        if rag_router_config.name:
            self.name = rag_router_config.name
        if rag_router_config.description:
            self.description = rag_router_config.description
        return self



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/3/12 19:36
# @Author  : heji
# @Email   : lc299034@antgroup.com
# @FileName: agent.py
"""The definition of agent paradigm."""
import json
from abc import abstractmethod, ABC
from datetime import datetime
from typing import Optional

from langchain_core.utils.json import parse_json_markdown

from agentuniverse.agent.agent_model import AgentModel
from agentuniverse.agent.input_object import InputObject
from agentuniverse.agent.output_object import OutputObject
from agentuniverse.agent.plan.planner.planner import Planner
from agentuniverse.agent.plan.planner.planner_manager import PlannerManager
from agentuniverse.base.annotation.trace import trace_agent
from agentuniverse.base.component.component_base import ComponentBase
from agentuniverse.base.component.component_enum import ComponentEnum
from agentuniverse.base.config.application_configer.application_config_manager \
    import ApplicationConfigManager
from agentuniverse.base.config.component_configer.configers.agent_configer \
    import AgentConfiger
from agentuniverse.base.util.logging.logging_util import LOGGER


class Agent(ComponentBase, ABC):
    """The parent class of all agent models, containing only attributes."""

    agent_model: Optional[AgentModel] = None

    def __init__(self):
        """Initialize the AgentModel with the given keyword arguments."""
        super().__init__(component_type=ComponentEnum.AGENT)

    @abstractmethod
    def input_keys(self) -> list[str]:
        """Return the input keys of the Agent."""
        pass

    @abstractmethod
    def output_keys(self) -> list[str]:
        """Return the output keys of the Agent."""
        pass

    @abstractmethod
    def parse_input(self, input_object: InputObject, agent_input: dict) -> dict:
        """Agent parameter parsing.

        Args:
            input_object (InputObject): input parameters passed by the user.
            agent_input (dict): agent input preparsed by the agent.
        Returns:
            dict: agent input parsed from `input_object` by the user.
        """
        pass

    @abstractmethod
    def parse_result(self, agent_result: dict) -> dict:
        """Agent result parser.

        Args:
            agent_result(dict): The raw result of the agent.
        Returns:
            dict: The parsed result of the agent
        """
        pass

    @trace_agent
    def run(self, **kwargs) -> OutputObject:
        """Agent instance running entry.

        Returns:
            OutputObject: Agent execution result
        """
        self.input_check(kwargs)
        input_object = InputObject(kwargs)

        agent_input = self.pre_parse_input(input_object)

        planner_result = self.execute(input_object, agent_input)

        agent_result = self.parse_result(planner_result)

        self.output_check(agent_result)
        output_object = OutputObject(agent_result)
        return output_object

    @trace_agent
    async def async_run(self, **kwargs) -> OutputObject:
        self.input_check(kwargs)
        input_object = InputObject(kwargs)

        agent_input = self.pre_parse_input(input_object)

        agent_result = await self.async_execute(input_object, agent_input)

        agent_result = self.parse_result(agent_result)

        self.output_check(agent_result)
        output_object = OutputObject(agent_result)
        return output_object

    def execute(self, input_object: InputObject, agent_input: dict) -> dict:
        """Execute agent instance.

        Args:
            input_object (InputObject): input parameters passed by the user.
            agent_input (dict): agent input parsed from `input_object` by the user.

        Returns:
            dict: planner result generated by the planner execution.
        """
        planner_base: Planner = PlannerManager().get_instance_obj(self.agent_model.plan.get('planner').get('name'))
        planner_result = planner_base.invoke(self.agent_model, agent_input, input_object)
        return planner_result

    async def async_execute(self, input_object: InputObject, agent_input: dict) -> dict:
        pass

    def pre_parse_input(self, input_object) -> dict:
        """Agent execution parameter pre-parsing.

        Args:
            input_object (InputObject): input parameters passed by the user.
        Returns:
            dict: agent input preparsed by the agent.
        """
        agent_input = dict()
        agent_input['chat_history'] = input_object.get_data('chat_history') or ''
        agent_input['background'] = input_object.get_data('background') or ''
        agent_input['image_urls'] = input_object.get_data('image_urls') or []
        agent_input['date'] = datetime.now().strftime('%Y-%m-%d')
        agent_input['session_id'] = input_object.get_data('session_id') or ''
        agent_input['agent_id'] = self.agent_model.info.get('name', '')
        self.parse_input(input_object, agent_input)
        return agent_input

    def get_instance_code(self) -> str:
        """Return the full name of the agent."""
        appname = ApplicationConfigManager().app_configer.base_info_appname
        name = self.agent_model.info.get('name')
        return (f'{appname}.'
                f'{self.component_type.value.lower()}.{name}')

    def input_check(self, kwargs: dict):
        """Agent parameter check."""
        for key in self.input_keys():
            if key not in kwargs.keys():
                raise Exception(f'Input must have key: {key}.')

    def output_check(self, kwargs: dict):
        """Agent result check."""
        if not isinstance(kwargs, dict):
            raise Exception('Output type must be dict.')
        for key in self.output_keys():
            if key not in kwargs.keys():
                raise Exception(f'Output must have key: {key}.')

    def initialize_by_component_configer(self, component_configer: AgentConfiger) -> 'Agent':
        """Initialize the Agent by the AgentConfiger object.

        Args:
            component_configer(AgentConfiger): the ComponentConfiger object
        Returns:
            Agent: the Agent object
        """
        agent_config: Optional[AgentConfiger] = component_configer.load()
        info: Optional[dict] = agent_config.info
        profile: Optional[dict] = agent_config.profile
        plan: Optional[dict] = agent_config.plan
        memory: Optional[dict] = agent_config.memory
        action: Optional[dict] = agent_config.action
        agent_model: Optional[AgentModel] = AgentModel(info=info, profile=profile,
                                                       plan=plan, memory=memory, action=action)
        self.agent_model = agent_model
        return self

    def langchain_run(self, input: str, callbacks=None, **kwargs):
        """Run the agent model using LangChain."""
        try:
            parse_result = parse_json_markdown(input)
        except Exception as e:
            LOGGER.error(f"langchain run parse_json_markdown error,input(parse_result) error({str(e)})")
            return "Error , Your Action Input is not a valid JSON string"
        output_object = self.run(**parse_result, callbacks=callbacks, **kwargs)
        result_dict = {}
        for key in self.output_keys():
            result_dict[key] = output_object.get_data(key)
        return result_dict

    def as_langchain_tool(self):
        """Convert to LangChain tool."""
        from langchain.agents.tools import Tool
        format_dict = {}
        for key in self.input_keys():
            format_dict.setdefault(key, "input val")
        format_str = json.dumps(format_dict)

        args_description = f"""
        to use this tool,your input must be a json string,must contain all keys of {self.input_keys()},
        and the value of the key must be a json string,the format of the json string is as follows:
        ```{format_str}```
        """
        return Tool(
            name=self.agent_model.info.get("name"),
            func=self.langchain_run,
            description=self.agent_model.info.get("description") + args_description
        )



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/10/10 18:53
# @Author  : wangchongshi
# @Email   : wangchongshi.wcs@antgroup.com
# @FileName: memory_storage.py
from typing import Optional, List

from agentuniverse.agent.memory.message import Message
from agentuniverse.base.component.component_base import ComponentBase
from agentuniverse.base.component.component_enum import ComponentEnum
from agentuniverse.base.config.component_configer.component_configer import ComponentConfiger


class MemoryStorage(ComponentBase):
    """The basic class for the memory storage.

    Attributes
        name (Optional[str]): The name of the memory storage class.
        description (Optional[str]): The description of the memory storage class.
    """

    name: Optional[str] = None
    description: Optional[str] = None
    component_type: ComponentEnum = ComponentEnum.MEMORY_STORAGE

    def _initialize_by_component_configer(self, memory_storage_config: ComponentConfiger) -> 'MemoryStorage':
        """Initialize the MemoryStorage by the ComponentConfiger object.

        Args:
            memory_storage_config(ComponentConfiger): A configer contains memory_storage basic info.
        Returns:
            MemoryStorage: A MemoryStorage instance.
        """
        if getattr(memory_storage_config, 'name', None):
            self.name = memory_storage_config.name
        if getattr(memory_storage_config, 'description', None):
            self.description = memory_storage_config.description
        return self

    def add(self, message_list: List[Message], session_id: str = None, agent_id: str = None, **kwargs) -> None:
        """Add messages to the memory db.

        Args:
            message_list (List[Message]): The list of messages to add.
            session_id (str): The session id of the memory to add.
            agent_id (str): The agent id of the memory to add.
        """
        pass

    def delete(self, session_id: str = None, agent_id: str = None, **kwargs) -> None:
        """Delete the memory from the database.

        Args:
            session_id (str): The session id of the memory to delete.
            agent_id (str): The agent id of the memory to delete.
        """
        pass

    def get(self, session_id: str = None, agent_id: str = None, top_k=10, **kwargs) -> List[Message]:
        """Get messages from the memory db.

        Args:
            session_id (str): The session id of the memory to get.
            agent_id (str): The agent id of the memory to get.
            top_k (int): The number of messages to return.
        Returns:
            List[Message]: A list of aU messages.
        """
        pass




from ..base.annotation.singleton import singleton
from ..base.component.component_enum import ComponentEnum
from ..base.component.component_manager_base import ComponentManagerBase
from .sqldb_wrapper import SQLDBWrapper


@singleton
class SQLDBWrapperManager(ComponentManagerBase[SQLDBWrapper]):
    """A singleton manager class of the DBWrapper."""

    def __init__(self):
        super().__init__(ComponentEnum.SQLDB_WRAPPER)




# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/3/26 15:17
# @Author  : fanen.lhy
# @Email   : fanen.lhy@antgroup.com
# @FileName: request_do.py

import datetime
from typing import Optional

from pydantic import BaseModel, Field


class RequestDO(BaseModel):
    """Data Object class of an agent service request."""

    id: Optional[int] = Field(description="ID", default=None)
    request_id: str = Field(description="Unique request id.")
    session_id: str = Field(description="Session id of the request.")
    query: str = Field(description="The query contents.")
    state: str = Field(description="State of the request.")
    result: dict = Field(description="Exec result.")
    steps: list = Field(description="Exec steps.")
    additional_args: dict = Field(description="Additional info.")
    gmt_create: Optional[datetime.datetime] = Field(
        description="Create time", default_factory=datetime.datetime.now)
    gmt_modified: Optional[datetime.datetime] = Field(
        description="Modified time", default_factory=datetime.datetime.now)



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/4/2 16:06
# @Author  : wangchongshi
# @Email   : wangchongshi.wcs@antgroup.com
# @FileName: llm_output.py
from typing import Any

from pydantic import BaseModel


class LLMOutput(BaseModel):
    """The basic class for llm output."""

    """The text of the llm output."""
    text: str

    """The raw data of the llm output."""
    raw: Any



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/3/18 14:32
# @Author  : wangchongshi
# @Email   : wangchongshi.wcs@antgroup.com
# @FileName: doc_parser.py
from typing import Union
from pathlib import Path
from typing import List, Optional, Dict

from agentuniverse.agent.action.knowledge.reader.reader import Reader
from agentuniverse.agent.action.knowledge.store.document import Document


class DocxReader(Reader):
    """Docx reader."""

    def _load_data(self, file: Union[str, Path], ext_info: Optional[Dict] = None) -> List[Document]:
        """Parse the docx file.

        Note:
            The docx file cannot be process in pagination.
            `docx2txt` is required to read DOCX files: `pip install docx2txt`
        """
        try:
            import docx2txt
        except ImportError:
            raise ImportError(
                "docx2txt is required to read Microsoft Word files: "
                "`pip install docx2txt`"
            )

        if isinstance(file, str):
            file = Path(file)
        text = docx2txt.process(file)
        metadata = {"file_name": file.name}
        if ext_info is not None:
            metadata.update(ext_info)

        return [Document(text=text, metadata=metadata or {})]



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/7/25 23:16
# @Author  : wangchongshi
# @Email   : wangchongshi.wcs@antgroup.com
# @FileName: tool_service.py
import os
from typing import List

from agentuniverse.agent.action.tool.tool import Tool
from agentuniverse.agent.action.tool.tool_manager import ToolManager
from agentuniverse.base.component.component_enum import ComponentEnum
from agentuniverse_product.base.product import Product
from agentuniverse_product.base.product_manager import ProductManager
from agentuniverse_product.base.util.yaml_util import write_yaml_file
from agentuniverse_product.service.model.tool_dto import ToolDTO
from agentuniverse_product.service.util.agent_util import register_product
from agentuniverse_product.service.util.common_util import get_core_path
from agentuniverse_product.service.util.tool_util import assemble_api_tool_config_data, \
    assemble_tool_product_config_data, register_tool, validate_create_api_tool_parameters


class ToolService:
    """Tool Service for aU-product."""

    @staticmethod
    def get_tool_list() -> List[ToolDTO]:
        """Get list of tools."""
        res = []
        product_list: List[Product] = ProductManager().get_instance_obj_list()
        if len(product_list) < 1:
            return res
        for product in product_list:
            if product.type == ComponentEnum.TOOL.value:
                tool_dto = ToolDTO(nickname=product.nickname, avatar=product.avatar, id=product.id)
                tool = product.instance
                tool_dto.description = tool.description
                tool_dto.parameters = tool.input_keys
                res.append(tool_dto)
        return res

    @staticmethod
    def get_tool_detail(tool_id: str) -> ToolDTO | None:
        """Get the tool detail.

        Args:
            tool_id (str): The id of the tool.

        Returns:
            ToolDTO | None: The tool detail.
        """
        if tool_id is None:
            return None
        tool: Tool = ToolManager().get_instance_obj(tool_id)
        if tool is None:
            return None
        product: Product = ProductManager().get_instance_obj(tool_id)
        tool_dto = ToolDTO(nickname=product.nickname if product else '',
                           avatar=product.avatar if product else '',
                           id=tool_id, description=tool.description, parameters=tool.input_keys)
        if hasattr(tool, 'openapi_spec'):
            tool_dto.openapi_schema = tool.openapi_spec
        return tool_dto

    @staticmethod
    def create_tool(tool_dto: ToolDTO) -> str:
        """Create a new tool instance.

        Args:
            tool_dto (ToolDTO): The tool DTO.

        Returns:
            str: The id of the tool.
        """
        # validate parameters
        validate_create_api_tool_parameters(tool_dto)

        # assemble product config data
        product_config_data = assemble_tool_product_config_data(tool_dto)

        # write product YAML file
        product_file_name = f"{tool_dto.id}_product"
        path = get_core_path()
        product_file_path = path / 'product' / 'tool' / f"{product_file_name}.yaml" if path \
            else os.path.join('..', '..', 'platform', 'difizen', 'product', 'tool', f"{product_file_name}.yaml")
        write_yaml_file(str(product_file_path), product_config_data)

        # assemble tool config data
        tool_config_data = assemble_api_tool_config_data(tool_dto)

        # write tool YAML file
        tool_file_path = path / 'tool' / f"{tool_dto.id}.yaml" if path \
            else os.path.join('..', '..', 'intelligence', 'agentic', 'tool', f"{tool_dto.id}.yaml")
        write_yaml_file(str(tool_file_path), tool_config_data)

        # register product and tool instance
        register_tool(str(tool_file_path))
        register_product(str(product_file_path))
        return tool_dto.id



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/9/29 17:11
# @Author  : wangchongshi
# @Email   : wangchongshi.wcs@antgroup.com
# @FileName: work_pattern.py
from abc import abstractmethod
from typing import Optional

from agentuniverse.agent.input_object import InputObject
from agentuniverse.base.component.component_base import ComponentBase
from agentuniverse.base.component.component_enum import ComponentEnum
from agentuniverse.base.config.component_configer.configers.work_pattern_configer import WorkPatternConfiger


class WorkPattern(ComponentBase):
    name: Optional[str] = None
    description: Optional[str] = None

    def __init__(self):
        """Initialize the ComponentBase."""
        super().__init__(component_type=ComponentEnum.WORK_PATTERN)

    @abstractmethod
    def invoke(self, input_object: InputObject, work_pattern_input: dict, **kwargs) -> dict:
        """Invoke the work pattern.

        Args:
            input_object (InputObject): The input parameters passed by the user.
            work_pattern_input (dict): Work pattern input dictionary.
            **kwargs: Additional keyword arguments.
        Returns:
            dict: The work pattern result.
        """
        pass

    @abstractmethod
    async def async_invoke(self, input_object: InputObject, work_pattern_input: dict, **kwargs) -> dict:
        """Asynchronously invoke the work pattern.

        Args:
            input_object (InputObject): The input parameters passed by the user.
            work_pattern_input (dict): Work pattern input dictionary.
            **kwargs: Additional keyword arguments.
        Returns:
            dict: The work pattern result.
        """
        pass

    def initialize_by_component_configer(self, work_pattern_configer: WorkPatternConfiger) -> 'WorkPattern':
        """Initialize the work pattern by the WorkPatternConfiger object.

        Args:
            work_pattern_configer(WorkPatternConfiger): the WorkPatternConfiger object
        Returns:
            WorkPattern: the work pattern object
        """
        self.name = work_pattern_configer.name
        self.description = work_pattern_configer.description
        return self

    def set_by_agent_model(self, **kwargs):
        pass



import json
from typing import Any, Union, AsyncIterator, Iterator, Optional, List, Sequence

import tiktoken
from langchain_core.language_models import BaseLanguageModel
from ollama import Options
from pydantic import Field

from agentuniverse.base.annotation.trace import trace_llm
from agentuniverse.base.config.component_configer.configers.llm_configer import LLMConfiger
from agentuniverse.base.util.env_util import get_from_env
from agentuniverse.llm.llm import LLM
from agentuniverse.llm.llm_output import LLMOutput
from agentuniverse.llm.ollama_langchain_instance import OllamaLangchainInstance


class OllamaLLM(LLM):
    base_url: Optional[str] = Field(
        default_factory=lambda: get_from_env("OLLAMA_BASE_URL") if get_from_env(
            "OLLAMA_BASE_URL") else "http://localhost:11434")
    """Base url the model is hosted under."""

    streaming: bool = True

    def _new_client(self):
        if self.client:
            return self.client
        from ollama import Client
        return Client(
            host=self.base_url,
        )

    def _new_async_client(self):
        if self.async_client:
            return self.async_client
        from ollama import AsyncClient
        return AsyncClient(
            host=self.base_url,
        )

    def _options(self):
        return Options(**{
            "num_ctx": self.max_context_length(),
            "num_predict": self.max_tokens,
            "temperature": self.temperature,
            "timeout": self.request_timeout,
            **(self.ext_info if self.ext_info else {}),
        })

    def _call(self, messages, stop=None, **kwargs) -> Union[LLMOutput, Iterator[LLMOutput]]:
        should_stream = kwargs.pop("stream", self.streaming)
        client = self._new_client()
        options = self._options()
        options.setdefault("stop", stop)
        res = client.chat(model=self.model_name, messages=messages, options=options, stream=should_stream)
        if should_stream:
            return self.generate_result(res)
        else:
            return LLMOutput(text=res.get("message").get('content'), raw=json.dumps(res))

    async def _acall(self, messages, stop=None, **kwargs) -> Union[LLMOutput, AsyncIterator[LLMOutput]]:
        client = self._new_async_client()
        should_stream = kwargs.pop("stream", self.streaming)
        options = self._options()
        options.setdefault("stop", stop)
        res = await client.chat(model=self.model_name, messages=messages, options=options, stream=should_stream)
        if not should_stream:
            return LLMOutput(text=res.get("message").get('content'), raw=json.dumps(res))
        if should_stream:
            return self.agenerate_result(res)

    def generate_result(self, data):
        for line in data:
            yield LLMOutput(text=line.get("message").get('content'), raw=json.dumps(line))

    async def agenerate_result(self, data):
        async for line in data:
            yield LLMOutput(text=line.get("message").get('content'), raw=json.dumps(line))

    def as_langchain(self) -> BaseLanguageModel:
        return OllamaLangchainInstance(
            self
        )

    def initialize_by_component_configer(self, component_configer: LLMConfiger) -> 'LLM':
        super().initialize_by_component_configer(component_configer)
        if 'base_url' in component_configer.configer.value:
            self.base_url = component_configer.configer.value['base_url']
        if 'max_context_length' in component_configer.configer.value:
            self._max_context_length = component_configer.configer.value['max_context_length']
        return self

    def get_num_tokens(self, text: str) -> int:
        try:
            encoding = tiktoken.encoding_for_model(self.model_name)
        except KeyError:
            encoding = tiktoken.get_encoding("cl100k_base")
        return len(encoding.encode(text))



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/4/2 15:27
# @Author  : jerry.zzw 
# @Email   : jerry.zzw@antgroup.com
# @FileName: agentuniverse.py
import importlib
import sys
import threading
from pathlib import Path

from agentuniverse.base.annotation.singleton import singleton
from agentuniverse.base.component.application_component_manager import ApplicationComponentManager
from agentuniverse.base.component.component_base import ComponentBase
from agentuniverse.base.config.application_configer.app_configer import AppConfiger
from agentuniverse.base.config.application_configer.application_config_manager import ApplicationConfigManager
from agentuniverse.base.config.component_configer.component_configer import ComponentConfiger
from agentuniverse.base.component.component_configer_util import ComponentConfigerUtil
from agentuniverse.base.config.config_type_enum import ConfigTypeEnum
from agentuniverse.base.config.configer import Configer
from agentuniverse.base.config.custom_configer.custom_key_configer import CustomKeyConfiger
from agentuniverse.base.component.component_enum import ComponentEnum
from agentuniverse.base.util.monitor.monitor import Monitor
from agentuniverse.base.util.system_util import get_project_root_path
from agentuniverse.base.util.logging.logging_util import init_loggers
from agentuniverse.agent_serve.web.request_task import RequestLibrary
from agentuniverse.agent_serve.web.rpc.grpc.grpc_server_booster import set_grpc_config
from agentuniverse.agent_serve.web.web_booster import ACTIVATE_OPTIONS
from agentuniverse.agent_serve.web.post_fork_queue import POST_FORK_QUEUE


@singleton
class AgentUniverse(object):
    """AgentUniverse framework object, responsible for the framework initialization,
       system variables management, etc."""

    def __init__(self):
        self.__application_container = ApplicationComponentManager()
        self.__config_container: ApplicationConfigManager = ApplicationConfigManager()
        self.__system_default_agent_package = ['agentuniverse.agent.default']
        self.__system_default_llm_package = ['agentuniverse.llm.default']
        self.__system_default_planner_package = ['agentuniverse.agent.plan.planner']
        self.__system_default_memory_package = ['agentuniverse.agent.memory.default']
        self.__system_default_prompt_package = ['agentuniverse.agent', 'agentuniverse.base.util']
        self.__system_default_embedding_package = ['agentuniverse.agent.action.knowledge.embedding']
        self.__system_default_doc_processor_package = ['agentuniverse.agent.action.knowledge.doc_processor']
        self.__system_default_reader_package = ['agentuniverse.agent.action.knowledge.reader.file']
        self.__system_default_rag_router_package = ['agentuniverse.agent.action.knowledge.rag_router']
        self.__system_default_query_paraphraser_package = ['agentuniverse.agent.action.knowledge.query_paraphraser']
        self.__system_default_memory_compressor_package = ['agentuniverse.agent.memory.memory_compressor']
        self.__system_default_memory_storage_package = ['agentuniverse.agent.memory.memory_storage']
        self.__system_default_work_pattern_package = ['agentuniverse.agent.work_pattern']

    def start(self, config_path: str = None, core_mode: bool = False):
        """Start the agentUniverse framework.

        """
        # get default config path
        project_root_path = get_project_root_path()
        sys.path.append(str(project_root_path.parent))
        self._add_to_sys_path(project_root_path, ['intelligence', 'app'])

        if not config_path:
            config_path = project_root_path / 'config' / 'config.toml'
            config_path = str(config_path)

        # load the configuration file
        configer = Configer(path=config_path).load()
        app_configer = AppConfiger().load_by_configer(configer)
        self.__config_container.app_configer = app_configer

        # load user custom key
        custom_key_configer_path = self.__parse_sub_config_path(
            configer.value.get('SUB_CONFIG_PATH', {}).get('custom_key_path'),
            config_path)
        CustomKeyConfiger(custom_key_configer_path)

        # init loguru loggers
        log_config_path = self.__parse_sub_config_path(
            configer.value.get('SUB_CONFIG_PATH', {}).get('log_config_path'),
            config_path)
        init_loggers(log_config_path)

        # init web request task database
        RequestLibrary(configer=configer)

        # Edit grpc config.
        grpc_activate = configer.value.get('GRPC', {}).get('activate')
        if grpc_activate and grpc_activate.lower() == 'true':
            ACTIVATE_OPTIONS["grpc"] = True
            set_grpc_config(configer)

        # Init gunicorn web server with config file.
        gunicorn_activate = configer.value.get('GUNICORN', {}).get('activate')
        if gunicorn_activate and gunicorn_activate.lower() == 'true':
            ACTIVATE_OPTIONS["gunicorn"] = True
            gunicorn_config_path = self.__parse_sub_config_path(
                configer.value.get('GUNICORN', {})
                .get('gunicorn_config_path'), config_path
            )
            from ..agent_serve.web.gunicorn_server import \
                GunicornApplication
            GunicornApplication(config_path=gunicorn_config_path)

        # init all extension module
        ext_classes = configer.value.get('EXTENSION_MODULES', {}).get('class_list')
        if isinstance(ext_classes, list):
            for ext_class in ext_classes:
                self.__dynamic_import_and_init(ext_class, configer)

        # init monitor module
        Monitor(configer=configer)

        # scan and register the components
        self.__scan_and_register(self.__config_container.app_configer)
        if core_mode:

            for _func, args, kwargs in POST_FORK_QUEUE:
                _func(*args, **kwargs)

    def __scan_and_register(self, app_configer: AppConfiger):
        """Scan the component directory and register the components.

        Args:
            app_configer(AppConfiger): the AppConfiger object
        """
        core_agent_package_list = ((app_configer.core_agent_package_list or app_configer.core_default_package_list)
                                   + self.__system_default_agent_package)
        core_knowledge_package_list = app_configer.core_knowledge_package_list or app_configer.core_default_package_list
        core_llm_package_list = ((app_configer.core_llm_package_list or app_configer.core_default_package_list)
                                 + self.__system_default_llm_package)
        core_planner_package_list = ((app_configer.core_planner_package_list or app_configer.core_default_package_list)
                                     + self.__system_default_planner_package)
        core_tool_package_list = app_configer.core_tool_package_list or app_configer.core_default_package_list
        core_service_package_list = app_configer.core_service_package_list or app_configer.core_default_package_list
        core_sqldb_wrapper_package_list = app_configer.core_sqldb_wrapper_package_list or app_configer.core_default_package_list
        core_memory_package_list = ((app_configer.core_memory_package_list or app_configer.core_default_package_list)
                                    + self.__system_default_memory_package)
        core_prompt_package_list = ((app_configer.core_prompt_package_list or app_configer.core_default_package_list)
                                    + self.__system_default_prompt_package)
        core_workflow_package_list = app_configer.core_workflow_package_list or app_configer.core_default_package_list
        core_embedding_package_list = ((app_configer.core_embedding_package_list or app_configer.core_default_package_list)
                                       + self.__system_default_embedding_package)
        core_doc_processor_package_list = ((app_configer.core_doc_processor_package_list or app_configer.core_default_package_list)
                                           + self.__system_default_doc_processor_package)
        core_reader_package_list = ((app_configer.core_reader_package_list or app_configer.core_default_package_list)
                                    + self.__system_default_reader_package)
        core_store_package_list = app_configer.core_store_package_list or app_configer.core_default_package_list
        core_rag_router_package_list = ((app_configer.core_rag_router_package_list or app_configer.core_default_package_list)
                                        + self.__system_default_rag_router_package)
        core_query_paraphraser_package_list = ((app_configer.core_query_paraphraser_package_list or app_configer.core_default_package_list)
                                               + self.__system_default_query_paraphraser_package)
        core_memory_compressor_package_list = ((app_configer.core_memory_compressor_package_list or app_configer.core_default_package_list)
                                               + self.__system_default_memory_compressor_package)
        core_memory_storage_package_list = ((app_configer.core_memory_storage_package_list or app_configer.core_default_package_list)
                                            + self.__system_default_memory_storage_package)
        core_work_pattern_package_list = ((app_configer.core_work_pattern_package_list or app_configer.core_default_package_list)
                                            + self.__system_default_work_pattern_package)

        component_package_map = {
            ComponentEnum.AGENT: core_agent_package_list,
            ComponentEnum.KNOWLEDGE: core_knowledge_package_list,
            ComponentEnum.LLM: core_llm_package_list,
            ComponentEnum.PLANNER: core_planner_package_list,
            ComponentEnum.TOOL: core_tool_package_list,
            ComponentEnum.SERVICE: core_service_package_list,
            ComponentEnum.SQLDB_WRAPPER: core_sqldb_wrapper_package_list,
            ComponentEnum.MEMORY: core_memory_package_list,
            ComponentEnum.PROMPT: core_prompt_package_list,
            ComponentEnum.WORKFLOW: core_workflow_package_list,
            ComponentEnum.EMBEDDING: core_embedding_package_list,
            ComponentEnum.DOC_PROCESSOR: core_doc_processor_package_list,
            ComponentEnum.READER: core_reader_package_list,
            ComponentEnum.STORE: core_store_package_list,
            ComponentEnum.RAG_ROUTER: core_rag_router_package_list,
            ComponentEnum.QUERY_PARAPHRASER: core_query_paraphraser_package_list,
            ComponentEnum.MEMORY_COMPRESSOR: core_memory_compressor_package_list,
            ComponentEnum.MEMORY_STORAGE: core_memory_storage_package_list,
            ComponentEnum.WORK_PATTERN: core_work_pattern_package_list
        }

        component_configer_list_map = {}
        for component_enum, package_list in component_package_map.items():
            if not package_list:
                continue
            component_configer_list = self.scan(package_list, ConfigTypeEnum.YAML, component_enum)
            component_configer_list_map[component_enum] = component_configer_list

        for component_enum, component_configer_list in component_configer_list_map.items():
            self.__register(component_enum, component_configer_list)

    def scan(self,
             package_list: [str],
             config_type_enum: ConfigTypeEnum,
             component_enum: ComponentEnum) -> list:
        """Scan the component directory and return certain component configer list.

        Args:
            package_list(list): the package list
            config_type_enum(ConfigTypeEnum): the configuration file type enumeration
            component_enum(ComponentEnum): the component enumeration

        Returns:
            list: the component configer list
        """
        component_configer_list = []
        for package_name in package_list:
            package_path = self.__package_name_to_path(package_name)
            path = Path(package_path)
            config_files = path.rglob(f'*.{config_type_enum.value}')
            for config_file in config_files:
                config_file_str = str(config_file)
                configer = Configer(path=config_file_str).load()
                component_configer = ComponentConfiger().load_by_configer(configer)
                component_config_type = component_configer.get_component_config_type()
                if component_config_type == component_enum.value:
                    component_configer_list.append(component_configer)
        return component_configer_list

    def __register(self, component_enum: ComponentEnum, component_configer_list: list[ComponentConfiger]):
        """Register the components.

        Args:
            component_enum(ComponentEnum): the component enumeration
            component_configer_list(list): the component configer list
        """
        component_manager_clz = ComponentConfigerUtil.get_component_manager_clz_by_type(component_enum)
        for component_configer in component_configer_list:
            configer_clz = ComponentConfigerUtil.get_component_config_clz_by_type(component_enum)
            configer_instance: ComponentConfiger = configer_clz().load_by_configer(component_configer.configer)
            component_clz = ComponentConfigerUtil.get_component_object_clz_by_component_configer(configer_instance)
            component_instance: ComponentBase = component_clz().initialize_by_component_configer(configer_instance)
            if component_instance is None:
                continue
            component_instance.component_config_path = component_configer.configer.path
            component_manager_clz().register(component_instance.get_instance_code(), component_instance)

    def __package_name_to_path(self, package_name: str) -> str:
        """Convert the package name to the package path.

        Args:
            package_name(str): the package name

        Returns:
            str: the package path
        """
        # get the package spec
        spec = importlib.util.find_spec(package_name)
        if spec is None:
            raise ImportError(f"Can not find {package_name}")
        # get the package path
        package_path = spec.submodule_search_locations[0] if spec.submodule_search_locations else spec.origin
        return package_path

    def __parse_sub_config_path(self, input_path: str,
                                reference_file_path: str) -> str | None:
        """Resolve a sub config file path according to main config file.

            Args:
                input_path(str): Absolute or relative path of sub config file.
                reference_file_path(str): Main config file path.
            Returns:
                str or None: A file path or none when no such file.
        """
        if not input_path:
            return None

        input_path_obj = Path(input_path)
        if input_path_obj.is_absolute():
            combined_path = input_path_obj
        else:
            reference_file_path_obj = Path(reference_file_path)
            combined_path = reference_file_path_obj.parent / input_path_obj

        return str(combined_path)

    def __dynamic_import_and_init(self, class_path: str, configer: Configer):
        """Resolve a sub config file path according to main config file.

            Args:
                class_path(str): Full class path like package_name.class_name.
                Auto read from config file.
        """

        module_path, _, class_name = class_path.rpartition('.')
        module = importlib.import_module(module_path)
        cls = getattr(module, class_name)
        cls(configer)

    def _add_to_sys_path(self, root_path, sub_dirs):
        for sub_dir in sub_dirs:
            app_path = root_path / sub_dir
            if app_path.exists():
                sys.path.append(str(app_path))



# !/usr/bin/env python3
# -*- coding:utf-8 -*-
import queue
# @Time    : 2024/4/2 16:04
# @Author  : wangchongshi
# @Email   : wangchongshi.wcs@antgroup.com
# @FileName: llm.py

from abc import abstractmethod
from typing import Optional, Any, AsyncIterator, Iterator, Union
from langchain_core.language_models.base import BaseLanguageModel
from langchain_core.runnables import Runnable

from agentuniverse.base.annotation.trace import trace_llm
from agentuniverse.base.component.component_base import ComponentBase
from agentuniverse.base.component.component_enum import ComponentEnum
from agentuniverse.base.config.application_configer.application_config_manager import ApplicationConfigManager
from agentuniverse.base.config.component_configer.configers.llm_configer import LLMConfiger
from agentuniverse.base.util.logging.logging_util import LOGGER
from agentuniverse.llm.llm_output import LLMOutput


class LLM(ComponentBase):
    """The basic class for llm model.

    Attributes:
        client (Any): The client of the llm.
        async_client (Any): The async client of the llm.
        name (Optional[str]): The name of the llm class.
        description (Optional[str]): The description of the llm model.
        model_name (Optional[str]): The name of the llm model, such as gpt-4, gpt-3.5-turbo.
        temperature (Optional[float]): The temperature of the llm model,
        what sampling temperature to use, between 0 and 2.
        request_timeout (Optional[int]): The request timeout for chat http requests.
        max_tokens (Optional[int]): The maximum number of [tokens](/tokenizer) that can be generated in the completion.
        streaming (Optional[bool]): Whether to stream the results or not.
        ext_info (Optional[dict]): The extended information of the llm model.
    """

    class Config:
        arbitrary_types_allowed = True

    client: Any = None
    async_client: Any = None
    name: Optional[str] = None
    description: Optional[str] = None
    model_name: Optional[str] = None
    temperature: Optional[float] = 0.5
    request_timeout: Optional[int] = None
    max_tokens: Optional[int] = 1024
    max_retries: Optional[int] = 2
    streaming: Optional[bool] = False
    ext_info: Optional[dict] = None
    tracing: Optional[bool] = None
    _max_context_length: Optional[int] = None
    langchain_instance: Optional[BaseLanguageModel] = None

    def __init__(self, **kwargs):
        """Initialize the llm."""
        super().__init__(component_type=ComponentEnum.LLM, **kwargs)

    def _new_client(self):
        """Initialize the client."""
        pass

    def _new_async_client(self):
        """Initialize the async client."""
        pass

    @abstractmethod
    def _call(self, *args: Any, **kwargs: Any) -> Union[LLMOutput, Iterator[LLMOutput]]:
        """Run the LLM."""

    @abstractmethod
    async def _acall(self, *args: Any, **kwargs: Any) -> Union[LLMOutput, AsyncIterator[LLMOutput]]:
        """Asynchronously run the LLM."""

    def as_langchain(self) -> BaseLanguageModel:
        """Convert to the langchain llm class."""
        pass

    def get_instance_code(self) -> str:
        """Return the full name of the llm."""
        appname = ApplicationConfigManager().app_configer.base_info_appname
        return f'{appname}.{self.component_type.value.lower()}.{self.name}'

    def initialize_by_component_configer(self, component_configer: LLMConfiger) -> 'LLM':
        """Initialize the LLM by the ComponentConfiger object.

        Args:
            component_configer(LLMConfiger): the ComponentConfiger object
        Returns:
            LLM: the LLM object
        """
        super().initialize_by_component_configer(component_configer)
        if component_configer.name:
            self.name = component_configer.name
        if component_configer.description:
            self.description = component_configer.description
        if component_configer.model_name:
            self.model_name = component_configer.model_name
        if component_configer.temperature:
            self.temperature = component_configer.temperature
        if component_configer.request_timeout:
            self.request_timeout = component_configer.request_timeout
        if component_configer.max_tokens:
            self.max_tokens = component_configer.max_tokens
        if component_configer.max_retries:
            self.max_retries = component_configer.max_retries
        if component_configer.streaming:
            self.streaming = component_configer.streaming
        if component_configer.ext_info:
            self.ext_info = component_configer.ext_info
        self.tracing = component_configer.tracing
        if 'max_context_length' in component_configer.configer.value:
            self._max_context_length = component_configer.configer.value['max_context_length']
        return self

    def set_by_agent_model(self, **kwargs):
        """ Assign values of parameters to the LLM model in the agent configuration."""
        # note: default shallow copy
        copied_obj = self.model_copy()
        if 'model_name' in kwargs and kwargs['model_name']:
            copied_obj.model_name = kwargs['model_name']
        if 'temperature' in kwargs and kwargs['temperature']:
            copied_obj.temperature = kwargs['temperature']
        if 'request_timeout' in kwargs and kwargs['request_timeout']:
            copied_obj.request_timeout = kwargs['request_timeout']
        if 'max_tokens' in kwargs and kwargs['max_tokens']:
            copied_obj.max_tokens = kwargs['max_tokens']
        if 'max_retries' in kwargs and kwargs['max_retries']:
            copied_obj.max_retries = kwargs['max_retries']
        if 'streaming' in kwargs and kwargs['streaming']:
            copied_obj.streaming = kwargs['streaming']
        if 'max_context_length' in kwargs and kwargs['max_context_length']:
            copied_obj._max_context_length = kwargs['max_context_length']
        return copied_obj

    def max_context_length(self) -> int:
        """Max context length.

        The total length of input tokens and generated tokens is limited by the model's context length.
        """
        return self._max_context_length

    @abstractmethod
    def get_num_tokens(self, text: str) -> int:
        """Get the number of tokens present in the text.

        Useful for checking if an input will fit in a model's context window.

        Args:
            text: The string input to tokenize.

        Returns:
            The integer number of tokens in the text.
        """

    def as_langchain_runnable(self, params=None) -> Runnable:
        """Get the langchain llm class."""
        if params is None:
            params = {}
        return self.as_langchain().bind(**params)

    @trace_llm
    def call(self, *args: Any, **kwargs: Any):
        """Run the LLM."""
        try:
            return self._call(*args, **kwargs)
        except Exception as e:
            LOGGER.error(f'Error in LLM call: {e}')
            raise e

    @trace_llm
    async def acall(self, *args: Any, **kwargs: Any):
        """Asynchronously run the LLM."""
        try:
            return await self._acall(*args, **kwargs)
        except Exception as e:
            LOGGER.error(f'Error in LLM acall: {e}')
            raise e



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/9/29 15:51
# @Author  : wangchongshi
# @Email   : wangchongshi.wcs@antgroup.com
# @FileName: agent_template.py
from abc import ABC
from typing import Optional, Any, List
from queue import Queue

from langchain_core.output_parsers import StrOutputParser
from langchain_core.runnables import RunnableSerializable

from agentuniverse.agent.action.knowledge.knowledge import Knowledge
from agentuniverse.agent.action.knowledge.knowledge_manager import KnowledgeManager
from agentuniverse.agent.action.knowledge.store.document import Document
from agentuniverse.agent.action.tool.tool import Tool
from agentuniverse.agent.action.tool.tool_manager import ToolManager
from agentuniverse.agent.agent import Agent
from agentuniverse.agent.input_object import InputObject
from agentuniverse.agent.memory.memory import Memory
from agentuniverse.agent.memory.memory_manager import MemoryManager
from agentuniverse.agent.memory.message import Message
from agentuniverse.base.config.component_configer.configers.agent_configer import AgentConfiger
from agentuniverse.base.util.agent_util import assemble_memory_input, assemble_memory_output
from agentuniverse.base.util.common_util import stream_output
from agentuniverse.base.util.memory_util import generate_messages
from agentuniverse.base.util.prompt_util import process_llm_token
from agentuniverse.llm.llm import LLM
from agentuniverse.llm.llm_manager import LLMManager
from agentuniverse.prompt.chat_prompt import ChatPrompt
from agentuniverse.prompt.prompt import Prompt
from agentuniverse.prompt.prompt_manager import PromptManager
from agentuniverse.prompt.prompt_model import AgentPromptModel


class AgentTemplate(Agent, ABC):
    llm_name: Optional[str] = ''
    memory_name: Optional[str] = None
    tool_names: Optional[list[str]] = None
    knowledge_names: Optional[list[str]] = None
    prompt_version: Optional[str] = None

    def execute(self, input_object: InputObject, agent_input: dict, **kwargs) -> dict:
        memory: Memory = self.process_memory(agent_input, **kwargs)
        llm: LLM = self.process_llm(**kwargs)
        prompt: Prompt = self.process_prompt(agent_input, **kwargs)
        return self.customized_execute(input_object, agent_input, memory, llm, prompt, **kwargs)

    async def async_execute(self, input_object: InputObject, agent_input: dict, **kwargs) -> dict:
        memory: Memory = self.process_memory(agent_input, **kwargs)
        llm: LLM = self.process_llm(**kwargs)
        prompt: Prompt = self.process_prompt(agent_input, **kwargs)
        return await self.customized_async_execute(input_object, agent_input, memory, llm, prompt, **kwargs)

    def customized_execute(self, input_object: InputObject, agent_input: dict, memory: Memory, llm: LLM, prompt: Prompt,
                           **kwargs) -> dict:
        assemble_memory_input(memory, agent_input)
        process_llm_token(llm, prompt.as_langchain(), self.agent_model.profile, agent_input)
        chain = prompt.as_langchain() | llm.as_langchain_runnable(
            self.agent_model.llm_params()) | StrOutputParser()
        res = self.invoke_chain(chain, agent_input, input_object, **kwargs)
        assemble_memory_output(memory=memory,
                               agent_input=agent_input,
                               content=f"Human: {agent_input.get('input')}, AI: {res}")
        self.add_output_stream(input_object.get_data('output_stream'), res)
        return {**agent_input, 'output': res}

    async def customized_async_execute(self, input_object: InputObject, agent_input: dict, memory: Memory,
                                       llm: LLM, prompt: Prompt, **kwargs) -> dict:
        assemble_memory_input(memory, agent_input)
        process_llm_token(llm, prompt.as_langchain(), self.agent_model.profile, agent_input)
        chain = prompt.as_langchain() | llm.as_langchain_runnable(
            self.agent_model.llm_params()) | StrOutputParser()
        res = await self.async_invoke_chain(chain, agent_input, input_object, **kwargs)
        assemble_memory_output(memory=memory,
                               agent_input=agent_input,
                               content=f"Human: {agent_input.get('input')}, AI: {res}")
        self.add_output_stream(input_object.get_data('output_stream'), res)
        return {**agent_input, 'output': res}

    def process_llm(self, **kwargs) -> LLM:
        return LLMManager().get_instance_obj(self.llm_name)

    def process_memory(self, agent_input: dict, **kwargs) -> Memory | None:
        memory: Memory = MemoryManager().get_instance_obj(component_instance_name=self.memory_name)
        if memory is None:
            return None

        chat_history: list = agent_input.get('chat_history')
        # generate a list of temporary messages from the given chat history and add them to the memory instance.
        temporary_messages: list[Message] = generate_messages(chat_history)
        if temporary_messages:
            memory.add(temporary_messages, **agent_input)

        params: dict = dict()
        params['agent_llm_name'] = self.llm_name
        return memory.set_by_agent_model(**params)

    def process_prompt(self, agent_input: dict, **kwargs) -> ChatPrompt:
        expert_framework = agent_input.pop('expert_framework', '') or ''

        profile: dict = self.agent_model.profile

        profile_instruction = profile.get('instruction')
        profile_instruction = expert_framework + profile_instruction if profile_instruction else profile_instruction

        profile_prompt_model: AgentPromptModel = AgentPromptModel(introduction=profile.get('introduction'),
                                                                  target=profile.get('target'),
                                                                  instruction=profile_instruction)

        # get the prompt by the prompt version
        version_prompt: Prompt = PromptManager().get_instance_obj(self.prompt_version)

        if version_prompt is None and not profile_prompt_model:
            raise Exception("Either the `prompt_version` or `introduction & target & instruction`"
                            " in agent profile configuration should be provided.")
        if version_prompt:
            version_prompt_model: AgentPromptModel = AgentPromptModel(
                introduction=getattr(version_prompt, 'introduction', ''),
                target=getattr(version_prompt, 'target', ''),
                instruction=expert_framework + getattr(version_prompt, 'instruction', ''))
            profile_prompt_model = profile_prompt_model + version_prompt_model

        chat_prompt = ChatPrompt().build_prompt(profile_prompt_model, ['introduction', 'target', 'instruction'])
        image_urls: list = agent_input.pop('image_urls', []) or []
        if image_urls:
            chat_prompt.generate_image_prompt(image_urls)
        return chat_prompt

    def invoke_chain(self, chain: RunnableSerializable[Any, str], agent_input: dict, input_object: InputObject,
                     **kwargs):
        if not input_object.get_data('output_stream'):
            res = chain.invoke(input=agent_input)
            return res
        result = []
        for token in chain.stream(input=agent_input):
            stream_output(input_object.get_data('output_stream', None), {
                'type': 'token',
                'data': {
                    'chunk': token,
                    'agent_info': self.agent_model.info
                }
            })
            result.append(token)
        return "".join(result)

    async def async_invoke_chain(self, chain: RunnableSerializable[Any, str], agent_input: dict,
                                 input_object: InputObject, **kwargs):
        if not input_object.get_data('output_stream'):
            res = await chain.ainvoke(input=agent_input)
            return res
        result = []
        async for token in chain.astream(input=agent_input):
            stream_output(input_object.get_data('output_stream', None), {
                'type': 'token',
                'data': {
                    'chunk': token,
                    'agent_info': self.agent_model.info
                }
            })
            result.append(token)
        return "".join(result)

    def invoke_tools(self, input_object: InputObject, **kwargs):
        if not self.tool_names:
            return ''

        tool_results: list = list()

        for tool_name in self.tool_names:
            tool: Tool = ToolManager().get_instance_obj(tool_name)
            if tool is None:
                continue
            tool_input = {key: input_object.get_data(key) for key in tool.input_keys}
            tool_results.append(str(tool.run(**tool_input)))
        return "\n\n".join(tool_results)

    def invoke_knowledge(self, query_str: str, input_object: InputObject, **kwargs):
        if not self.knowledge_names or not query_str:
            return ''

        knowledge_results: list = list()

        for knowledge_name in self.knowledge_names:
            knowledge: Knowledge = KnowledgeManager().get_instance_obj(knowledge_name)
            if knowledge is None:
                continue
            knowledge_res: List[Document] = knowledge.query_knowledge(
                query_str=query_str,
                **input_object.to_dict()
            )
            knowledge_results.append(knowledge.to_llm(knowledge_res))
        return "\n\n".join(knowledge_results)

    def validate_required_params(self):
        pass

    def add_output_stream(self, output_stream: Queue, agent_output: str) -> None:
        pass

    def initialize_by_component_configer(self, component_configer: AgentConfiger) -> 'AgentTemplate':
        super().initialize_by_component_configer(component_configer)
        self.llm_name = self.agent_model.profile.get('llm_model', {}).get('name')
        self.memory_name = self.agent_model.memory.get('name')
        self.tool_names = self.agent_model.action.get('tool', [])
        self.knowledge_names = self.agent_model.action.get('knowledge', [])
        return self



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/5/21 17:13
# @Author  : fanen.lhy
# @Email   : fanen.lhy@antgroup.com
# @FileName: sqldb_wrapper_config.py

from typing import Optional

from agentuniverse.base.config.component_configer.component_configer import ComponentConfiger
from agentuniverse.base.config.configer import Configer
from agentuniverse.base.util.system_util import parse_dynamic_str


class SQLDBWrapperConfiger(ComponentConfiger):
    """The SQLDBWrapperConfiger class, used to load and manage the sql db wrapper
    configuration."""

    _ComponentConfiger__metadata_class: Optional[str] = None
    _ComponentConfiger__metadata_module: Optional[str] = None

    def __init__(self, configer: Optional[Configer] = None):
        """Initialize the DBWrapperConfiger."""
        super().__init__(configer)
        self.__name: Optional[str] = None
        self.__description: Optional[str] = None

        self.sql_database_args: dict = {}
        self.db_uri: Optional[str] = None
        self.engine_args: dict = {}

        self.__set_default_meta_info()

    @property
    def name(self) -> Optional[str]:
        """Name field."""
        return self.__name

    @property
    def description(self) -> Optional[str]:
        """Description field."""
        return self.__description

    def __set_default_meta_info(self):
        """Set default instantiated class of service."""
        if (not hasattr(self, '_ComponentConfiger__metadata_module')
                or self._ComponentConfiger__metadata_module is None):
            self._ComponentConfiger__metadata_module = (
                "agentuniverse.database.sqldb_wrapper"
            )
        if (not hasattr(self, '_ComponentConfiger__metadata_class')
                or self._ComponentConfiger__metadata_class is None):
            self._ComponentConfiger__metadata_class = 'SQLDBWrapper'

    def load(self) -> 'SQLDBWrapperConfiger':
        """Setting property using own configer member property.

        Returns:
            SQLDBWrapperConfiger: A SQLDBWrapperConfiger instance.
        """
        return self.load_by_configer(self.configer)

    def load_by_configer(self, configer: Configer) -> 'SQLDBWrapperConfiger':
        """Initialize self using given configer, get SQLDBWrapperConfiger property
        from it.
        Args:
            configer(Configer): A Configer instance.
        Returns:
            SQLDBWrapperConfiger: A SQLDBWrapperConfiger instance.
        """
        super().load_by_configer(configer)
        self.__set_default_meta_info()
        try:
            self.__name = configer.value.get('name')
            self.__description = configer.value.get('description')
            self.db_uri = parse_dynamic_str(configer.value.get('db_uri'))
            self.engine_args = dict(configer.value.get('engine_args', {}))
            self.sql_database_args = dict(configer.value.get('sql_database_args', {}))

        except Exception as e:
            raise Exception(f"Failed to parse the db wrapper configuration: {e}")
        return self



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/7/26 14:20
# @Author  : wangchongshi
# @Email   : wangchongshi.wcs@antgroup.com
# @FileName: session_service.py
import uuid
from datetime import datetime
from typing import List

from agentuniverse_product.dal.message_library import MessageLibrary
from agentuniverse_product.dal.model.message_do import MessageDO
from agentuniverse_product.dal.model.session_do import SessionDO
from agentuniverse_product.dal.session_library import SessionLibrary
from agentuniverse_product.service.model.message_dto import MessageDTO
from agentuniverse_product.service.model.session_dto import SessionDTO


class SessionService:
    """Session Service for aU-product."""

    @staticmethod
    def create_session(agent_id: str) -> str:
        """Create a new session for a given agent id."""
        if agent_id is None:
            raise ValueError("agent_id is required parameter.")
        session_id = str(uuid.uuid4())
        session_do = SessionDO(
            session_id=session_id,
            agent_id=agent_id,
            ext_info=dict(),
        )
        return SessionLibrary().add_session(session_do)

    @staticmethod
    def update_session(session_id: str, agent_id: str, update_time: datetime) -> str:
        """Update a session."""
        return SessionLibrary().update_session(
            SessionDO(session_id=session_id, agent_id=agent_id, gmt_modified=update_time))

    @staticmethod
    def delete_session(session_id: str) -> str:
        """Delete the session with the specified session id."""
        if session_id is None:
            raise ValueError("session_id is required parameter.")
        session_id = SessionLibrary().delete_session(session_id)
        if session_id:
            MessageLibrary().delete_messages(session_id)
        return session_id or ''

    @staticmethod
    def get_session_list(agent_id: str) -> List[SessionDTO]:
        """Get a list of sessions for a given agent id."""
        if agent_id is None:
            raise ValueError("agent_id is required parameter.")
        session_do_list: List[SessionDO] = SessionLibrary().get_session_list(agent_id)
        if len(session_do_list) == 0:
            return []
        # newest 10 sessions
        session_do_list = session_do_list[:10]
        session_messages_map = {}
        for session_do in session_do_list:
            message_do_list: List[MessageDO] = MessageLibrary().get_messages(session_do.session_id)
            session_messages_map[session_do.session_id] = message_do_list
        return SessionService().convert_to_session_dto(session_do_list, session_messages_map)

    @staticmethod
    def get_session_detail(id: str, top_k: int = None) -> SessionDTO | None:
        """Get the session detail for a given session id."""
        if id is None:
            raise ValueError("Session id is required parameter.")
        session_do: SessionDO = SessionLibrary().get_session_detail(id)
        if session_do is None:
            return session_do
        # newest top k messages
        message_do_list: List[MessageDO] = MessageLibrary().get_messages(session_do.session_id)
        if top_k:
            message_do_list = message_do_list[-top_k:]
        return SessionService().convert_to_session_dto([session_do], {session_do.session_id: message_do_list})[0]

    @staticmethod
    def convert_to_session_dto(session_do_list: List[SessionDO], session_messages_map: dict[str, List[MessageDO]]) -> \
            List[SessionDTO]:
        """Convert the given session do list to session dto list."""
        res = []
        if len(session_do_list) == 0:
            return res
        for session_do in session_do_list:
            session_dto = SessionDTO(id=session_do.session_id, agent_id=session_do.agent_id,
                                     gmt_created=session_do.gmt_created.strftime('%Y-%m-%d %H:%M:%S'),
                                     gmt_modified=session_do.gmt_modified.strftime('%Y-%m-%d %H:%M:%S'))
            message_dto_list = []
            message_do_list = session_messages_map[session_do.session_id]
            for message_do in message_do_list:
                message_dto = MessageDTO(id=message_do.id, session_id=session_do.session_id,
                                         gmt_created=message_do.gmt_created.strftime('%Y-%m-%d %H:%M:%S'),
                                         gmt_modified=message_do.gmt_modified.strftime('%Y-%m-%d %H:%M:%S'),
                                         content=message_do.content)
                message_dto_list.append(message_dto)
            session_dto.messages = message_dto_list
            res.append(session_dto)
        return res



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/7/24 16:34
# @Author  : wangchongshi
# @Email   : wangchongshi.wcs@antgroup.com
# @FileName: product_manager.py
from agentuniverse.base.annotation.singleton import singleton
from agentuniverse.base.component.component_enum import ComponentEnum
from agentuniverse.base.component.component_manager_base import ComponentManagerBase
from agentuniverse_product.base.product import Product


@singleton
class ProductManager(ComponentManagerBase[Product]):
    """A singleton manager class of the product."""

    def __init__(self):
        super().__init__(ComponentEnum.PRODUCT)



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/10/24 21:12
# @Author  : wangchongshi
# @Email   : wangchongshi.wcs@antgroup.com
# @FileName: common_util.py
from queue import Queue


def stream_output(output_stream: Queue, data: dict):
    """Add data to the output stream.

    Args:
        output_stream (Queue): The output stream.
        data (dict): The data to be streamed.
    """
    if output_stream is None:
        return
    output_stream.put_nowait(data)



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/8/5 15:48
# @Author  : fanen.lhy
# @Email   : fanen.lhy@antgroup.com
# @FileName: dashscope_reranker.py

from typing import List, Optional
import dashscope
from http import HTTPStatus

from agentuniverse.agent.action.knowledge.doc_processor.doc_processor import \
    DocProcessor
from agentuniverse.agent.action.knowledge.store.document import Document
from agentuniverse.agent.action.knowledge.store.query import Query
from agentuniverse.base.config.component_configer.component_configer import \
    ComponentConfiger

MODEL_NAME_MAP = {
    "gte_rerank": dashscope.TextReRank.Models.gte_rerank
}


class DashscopeReranker(DocProcessor):
    model_name: str = "gte_rerank"
    top_n: int = 10

    def _process_docs(self, origin_docs: List[Document], query: Query = None) -> \
            List[Document]:
        if not query or not query.query_str:
            raise Exception("Dashscope reranker need an origin string query.")
        if len(origin_docs)<1:
            return origin_docs
        documents_texts = []
        for _doc in origin_docs:
            documents_texts.append(_doc.text)
        resp = dashscope.TextReRank.call(
            model=MODEL_NAME_MAP.get(self.model_name),
            query=query.query_str,
            documents=documents_texts,
            top_n=self.top_n,
            return_documents=False
        )
        if resp.status_code == HTTPStatus.OK:
            results = resp.output.results
        else:
            raise Exception(f"Dashscope rerank api call error: {resp}")
        rerank_docs = []
        for _result in results:
            index = _result.index
            if origin_docs[index].metadata:
                origin_docs[index].metadata["relevance_score"] = _result.relevance_score
            else:
                origin_docs[index].metadata = {"relevance_score": _result.relevance_score}
            rerank_docs.append(origin_docs[index])

        return rerank_docs

    def _initialize_by_component_configer(self,
                                         doc_processor_configer: ComponentConfiger) -> 'DocProcessor':
        super()._initialize_by_component_configer(doc_processor_configer)
        if hasattr(doc_processor_configer, "model_name"):
            self.model_name = doc_processor_configer.model_name
        if hasattr(doc_processor_configer, "top_n"):
            self.top_n = doc_processor_configer.top_n
        return self



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/3/14 12:08
# @Author  : jerry.zzw 
# @Email   : jerry.zzw@antgroup.com
# @FileName: component_manager.py
from agentuniverse.agent.action.knowledge.knowledge_manager import KnowledgeManager
from agentuniverse.agent.action.tool.tool_manager import ToolManager
from agentuniverse.agent.plan.planner.planner_manager import PlannerManager
from agentuniverse.agent.agent_manager import AgentManager
from agentuniverse.base.annotation.singleton import singleton
from agentuniverse.llm.llm_manager import LLMManager


@singleton
class ApplicationComponentManager(object):
    """The ComponentManager class, which is used to manage the components."""

    def __init__(self):
        """Initialize the ComponentManager."""
        self.__agent_manager: AgentManager = AgentManager()
        self.__llm_manager: LLMManager = LLMManager()
        self.__planner_manager: PlannerManager = PlannerManager()
        self.__knowledge: KnowledgeManager = KnowledgeManager()
        self.__tool_manager: ToolManager = ToolManager()



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/8/23 16:12
# @Author  : wangchongshi
# @Email   : wangchongshi.wcs@antgroup.com
# @FileName: workflow_dto.py
from typing import Optional

from pydantic import BaseModel, Field


class WorkflowDTO(BaseModel):
    id: Optional[str] = Field(description="ID", default=None)
    name: Optional[str] = Field(description="workflow name", default="")
    description: Optional[str] = Field(description="workflow description", default="")
    graph: Optional[dict] = Field(description="workflow graph config", default=None)



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/3/18 19:18
# @Author  : wangchongshi
# @Email   : wangchongshi.wcs@antgroup.com
# @FileName: embedding.py
from abc import abstractmethod
from typing import List, Optional

from langchain_core.embeddings import Embeddings as LCEmbeddings

from agentuniverse.base.component.component_base import ComponentEnum
from agentuniverse.base.component.component_base import ComponentBase
from agentuniverse.base.config.component_configer.component_configer import \
    ComponentConfiger


class Embedding(ComponentBase):
    """The basic class for embedding.

    Attributes:
        embedding_model_name (Optional[str]): The name of the embedding model.
    """

    component_type: ComponentEnum = ComponentEnum.EMBEDDING
    name: Optional[str] = None
    description: Optional[str] = None
    embedding_model_name: Optional[str] = None
    embedding_dims: Optional[int] = None

    @abstractmethod
    def get_embeddings(self, text: List[str], **kwargs) -> List[List[float]]:
        """Get embeddings."""

    @abstractmethod
    async def async_get_embeddings(self, texts: List[str], **kwargs) -> List[
        List[float]]:
        """Asynchronously get embeddings."""

    def as_langchain(self) -> LCEmbeddings:
        """Convert the agentUniverse(aU) embedding class to the langchain embedding class."""
        pass

    def _initialize_by_component_configer(self,
                                          embedding_configer: ComponentConfiger) \
            -> 'Embedding':
        """Initialize the embedding by the ComponentConfiger object.

        Args:
            embedding_configer(ComponentConfiger): A configer contains embedding
            basic info.
        Returns:
            Embedding: A embedding instance.
        """
        if embedding_configer.name:
            self.name = embedding_configer.name
        if embedding_configer.description:
            self.description = embedding_configer.description
        if hasattr(embedding_configer, "embedding_dims"):
            self.embedding_dims = embedding_configer.embedding_dims
        if hasattr(embedding_configer, "embedding_model_name"):
            self.embedding_model_name = embedding_configer.embedding_model_name
        return self



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/8/23 18:01
# @Author  : wangchongshi
# @Email   : wangchongshi.wcs@antgroup.com
# @FileName: condition_node.py
from typing import List, Optional

from agentuniverse.workflow.node.enum import NodeEnum, ConditionComparisonEnum, NodeStatusEnum
from agentuniverse.workflow.node.node import Node, NodeData
from agentuniverse.workflow.node.node_config import ConditionNodeInputParams, ConditionBranchParams, ConditionParams, \
    NodeInputParams, NodeOutputParams
from agentuniverse.workflow.node.node_output import NodeOutput
from agentuniverse.workflow.workflow_output import WorkflowOutput


class ConditionNodeData(NodeData):
    inputs: Optional[ConditionNodeInputParams] = None


class ConditionNode(Node):
    """The basic class of the condition node."""
    _data_cls = ConditionNodeData

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.type = NodeEnum.CONDITION

    def _run(self, workflow_output: WorkflowOutput) -> NodeOutput:
        inputs: ConditionNodeInputParams = self._data.inputs
        condition_branches: List[ConditionBranchParams] = inputs.branches
        condition_branch: ConditionBranchParams = condition_branches[0]
        condition: ConditionParams = condition_branch.conditions[0]

        def resolve_value(node_input: NodeInputParams):
            if node_input.value.type == 'reference':
                reference_node_id = node_input.value.content[0]
                reference_output_params: List[NodeOutputParams] = workflow_output.workflow_parameters.get(
                    reference_node_id, [])
                return next(
                    (param.value for param in reference_output_params if param.name == node_input.value.content[1]),
                    None)
            return node_input.value.content

        left_val = resolve_value(condition.left)
        right_val = resolve_value(condition.right) if condition.right else None

        compare: str = condition.compare
        res = False

        if compare == ConditionComparisonEnum.EQUAL.value:
            res = left_val == right_val
        elif compare == ConditionComparisonEnum.NOT_EQUAL.value:
            res = left_val != right_val
        elif compare == ConditionComparisonEnum.BLANK.value:
            res = left_val is None
        return NodeOutput(
            node_id=self.id,
            status=NodeStatusEnum.SUCCEEDED,
            edge_source_handler=condition_branch.name if res else 'branch-default'
        )



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/8/20 22:10
# @Author  : wangchongshi
# @Email   : wangchongshi.wcs@antgroup.com
# @FileName: agent_node.py
import re
from typing import List, Dict

from agentuniverse.agent.agent import Agent
from agentuniverse.agent.agent_manager import AgentManager
from agentuniverse.agent.output_object import OutputObject
from agentuniverse.workflow.node.enum import NodeEnum, NodeStatusEnum
from agentuniverse.workflow.node.node import Node, NodeData
from agentuniverse.workflow.node.node_config import AgentNodeInputParams, NodeInfoParams, NodeOutputParams
from agentuniverse.workflow.node.node_output import NodeOutput
from agentuniverse.workflow.workflow_output import WorkflowOutput


class AgentNodeData(NodeData):
    inputs: AgentNodeInputParams


class AgentNode(Node):
    """The basic class of the agent node."""

    _data_cls = AgentNodeData

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.type = NodeEnum.AGENT

    def _run(self, workflow_output: WorkflowOutput) -> NodeOutput:
        inputs: AgentNodeInputParams = self._data.inputs

        param_map = {
            'prompt': None,
            'id': None
        }

        for agent_param in inputs.agent_param:
            if agent_param.name in param_map:
                if isinstance(agent_param.value, str):
                    param_map[agent_param.name] = agent_param.value
                else:
                    param_map[agent_param.name] = agent_param.value.get('content', None)

        agent_id = param_map['id']
        prompt = param_map['prompt']

        agent: Agent = AgentManager().get_instance_obj(agent_id)
        if agent is None:
            raise ValueError("No agent with id {} was found.".format(agent_id))

        # Extract variables from the prompt template
        template_variables = re.findall(r'\{\{(.*?)\}\}', prompt)

        agent_input_params = self._resolve_input_params(inputs.input_param, workflow_output)

        # Replace variables in the prompt
        try:
            for var in template_variables:
                if var not in agent_input_params:
                    raise KeyError(f"The variable '{var}' is not found in the input params.")
                prompt = prompt.replace(f'{{{{{var}}}}}',
                                        str(agent_input_params[var]) if agent_input_params[var] else '')
        except KeyError as e:
            raise ValueError(f"Error processing template variables: {e}")

        agent_output: OutputObject = agent.run(input=prompt)
        agent_output_dict = agent_output.to_dict()
        output_params: List[NodeOutputParams] = self._data.outputs

        for output_param in output_params:
            output_param.value = agent_output_dict.get(output_param.name, None)
        workflow_output.workflow_parameters[self.id] = output_params
        return NodeOutput(node_id=self.id, status=NodeStatusEnum.SUCCEEDED, result=output_params)



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/7/23 14:00
# @Author  : fanen.lhy
# @Email   : fanen.lhy@antgroup.com
# @FileName: doc_processor.py

from abc import abstractmethod
from typing import List, Optional

from agentuniverse.agent.action.knowledge.store.query import Query
from agentuniverse.agent.action.knowledge.store.document import Document
from agentuniverse.base.component.component_base import ComponentEnum
from agentuniverse.base.component.component_base import ComponentBase
from agentuniverse.base.config.component_configer.component_configer import \
    ComponentConfiger


class DocProcessor(ComponentBase):
    """The basic class for doc processor.
    """

    component_type: ComponentEnum = ComponentEnum.DOC_PROCESSOR
    name: Optional[str] = None
    description: Optional[str] = None

    class Config:
        arbitrary_types_allowed = True

    def process_docs(self, origin_docs: List[Document], query: Query = None) -> \
            List[Document]:
        """Process input documents，return should also be a document list."""
        return self._process_docs(origin_docs, query)

    @abstractmethod
    def _process_docs(self, origin_docs: List[Document],
                      query: Query = None) -> \
            List[Document]:
        """Process input documents，return should also be a document list."""
        pass

    def _initialize_by_component_configer(self,
                                         doc_processor_configer: ComponentConfiger) \
            -> 'DocProcessor':
        """Initialize the DocProcessor by the ComponentConfiger object.

        Args:
            doc_processor_configer(ComponentConfiger): A configer contains DocProcessor
            basic info.
        Returns:
            DocProcessor: A DocProcessor instance.
        """
        self.name = doc_processor_configer.name
        self.description = doc_processor_configer.description
        return self



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/8/6 10:44
# @Author  : fanen.lhy
# @Email   : fanen.lhy@antgroup.com
# @FileName: jieba_keyword_extractor.py
from typing import List

from agentuniverse.agent.action.knowledge.doc_processor.doc_processor import \
    DocProcessor
from agentuniverse.agent.action.knowledge.store.document import Document
from agentuniverse.agent.action.knowledge.store.query import Query
from agentuniverse.base.config.component_configer.component_configer import \
    ComponentConfiger

import jieba
import jieba.analyse

# nltk english stopwords
stop_words = {'or', "mustn't", 'how', 'their', 'again', 'few', 'other', 'who',
              'being', 'theirs', 'during', 'if', 'on', 'she', 'wouldn', 'why',
              'above', 'll', "weren't", 'your', 'are', 'an', 'over', 'his',
              'hasn', 'off', 'you', 'he', 'was', "you'd", 'me', 'ain', 'any',
              'what', 'most', 're', 'haven', "isn't", 'there', "it's", 'same',
              'm', 'only', 'my', 'needn', 'too', 'into', 'in', 'by', 'between',
              "that'll", "mightn't", "aren't", 'am', 'up', 'having', "you'll",
              "you're", 'these', 'mustn', 'himself', 'down', 'such', 'wasn',
              'ourselves', 'did', 'because', 'should', 'won', 'about', 'aren',
              'don', 'while', 't', 'isn', 'have', 'whom', 'myself', 'itself',
              'this', 'will', 'and', 'further', 'no', 'where', 'ma', 'yours',
              'been', "didn't", 'that', 'had', 'when', 'we', 'herself', 'some',
              'has', "she's", "needn't", "should've", 'of', "won't", 'both',
              'which', "haven't", 'yourself', 'through', 'the', 'from',
              "you've", 'for', 'then', 'hadn', 'a', 'them', 'as', 'after',
              'themselves', "shouldn't", 'they', 'y', 'doesn', 'didn', 'here',
              'ours', 'own', 'it', "hadn't", 'each', 'our', 'shouldn', 'all',
              'out', 'before', 'couldn', 'd', "doesn't", 'hers', "hasn't",
              'than', 'at', "don't", 'not', 'to', 'is', 'with', 'until',
              'does', 'yourselves', 'under', 'below', 'i', 'those', "wouldn't",
              'once', "couldn't", 'just', 's', 'shan', "wasn't", 'him', 'so',
              'can', 'doing', 'o', 'her', 'were', 'now', 'very', 'weren',
              'its', 'against', 'do', 've', 'be', 'mightn', 'but', "shan't",
              'nor', 'more'}

chinese_stopwords = {'的', '了', '在', '是', '我', '有', '和', '就', '不',
                     '人', '都', '一', '一个', '上', '也', '很', '到', '说',
                     '要', '去', '你', '会', '着', '没有', '看', '好', '自己',
                     '请问', '您', '他', '于', '及', '即', '为', '最', '从', '以',
                     '了', '将', '与', '吗', '吧', '中', '#', '什么', '怎么', '哪个',
                     '哪些', '啥', '相关'}


class JiebaKeywordExtractor(DocProcessor):
    top_k: int = 3

    def _process_docs(self, origin_docs: List[Document], query: Query = None) \
            -> List[Document]:
        for _doc in origin_docs:
            words = jieba.lcut(_doc.text)
            filtered_words = [word for word in words if word not in
                              chinese_stopwords and word.lower() not in stop_words]
            keywords = jieba.analyse.extract_tags(" ".join(filtered_words),
                                                  topK=self.top_k)
            _doc.keywords.update(keywords)

        return origin_docs

    def _initialize_by_component_configer(self,
                                         doc_processor_configer: ComponentConfiger) -> 'DocProcessor':
        super()._initialize_by_component_configer(doc_processor_configer)
        if hasattr(doc_processor_configer, "top_k"):
            self.top_k = doc_processor_configer.top_k
        return self
